"""PvE Views"""

import logging
from datetime import datetime

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import redirect, render

# Django
from django.utils.translation import gettext_lazy as _

from allianceauth.eveonline.models import EveCorporationInfo

from ledger.api.helpers import get_all_corporations_from_alliance, get_alliance
from ledger.view_helpers.core import add_info_to_context

logger = logging.getLogger(__name__)


@login_required
@permission_required("ledger.advanced_access")
def alliance_ledger_index(request):
    """Alliance Ledger Index View"""
    context = {}
    context = add_info_to_context(request, context)
    try:
        alliance_id = request.user.profile.main_character.alliance_id
        if alliance_id is None:
            raise AttributeError
    except AttributeError:
        messages.error(request, "You do not have an alliance.")
        return redirect("ledger:index")
    return redirect("ledger:alliance_ledger", alliance_id)


@login_required
@permission_required("ledger.advanced_access")
def alliance_ledger(request, alliance_id):
    """
    Alliance Ledger
    """
    # pylint: disable=duplicate-code
    current_year = datetime.now().year
    years = [current_year - i for i in range(6)]

    context = {
        "title": "Alliance Ledger",
        "years": years,
        "alliance_id": alliance_id,
        "entity_pk": alliance_id,
        "entity_type": "alliance",
    }
    context = add_info_to_context(request, context)
    return render(request, "ledger/allyledger/alliance_ledger.html", context=context)


@login_required
@permission_required("ledger.advanced_access")
def alliance_overview(request):
    """
    Alliance Overview
    """
    context = {
        "title": "Alliance Overview",
    }
    context = add_info_to_context(request, context)
    return render(
        request, "ledger/allyledger/admin/alliance_overview.html", context=context
    )


@login_required
@permission_required("ledger.admin_access")
def alliance_administration(request, alliance_id):
    """
    Alliance Administration
    """
    perm, alliance = get_alliance(request, alliance_id)

    if perm is False:
        msg = _("Permission Denied")
        messages.error(request, msg)
        return redirect("ledger:alliance_ledger_index")
    if perm is None:
        msg = _("Alliance not Found")
        messages.info(request, msg)
        return redirect("ledger:alliance_ledger_index")

    corporations = get_all_corporations_from_alliance(request, alliance.alliance_id)[1]
    all_corporations = EveCorporationInfo.objects.filter(
        alliance__alliance_id=alliance_id
    ).order_by("corporation_name")

    corp_audit_ids = corporations.values_list("corporation__corporation_id", flat=True)
    missing_corporations = all_corporations.exclude(
        corporation_id__in=corp_audit_ids
    ).order_by("corporation_name")

    context = {
        "alliance_id": alliance_id,
        "title": "Alliance Administration",
        "corporations": corporations,
        "missing_corporations": missing_corporations,
    }
    context = add_info_to_context(request, context)
    return render(
        request,
        "ledger/allyledger/admin/alliance_administration.html",
        context=context,
    )



import logging

from django.db import models
from django.db.models import DecimalField, F, Q, Sum, Value
from django.db.models.functions import Coalesce

from ledger.constants import (
    ASSETS,
    BOUNTY_PRIZES,
    CONTRACT,
    DAILY_GOAL_REWARD,
    DONATION,
    ESS_TRANSFER,
    INCURSION,
    INSURANCE,
    LP,
    MARKET,
    MILESTONE_REWARD,
    MISSION_REWARD,
    PLANETARY,
    PRODUCTION,
    SKILL,
    TRAVELING,
)

logger = logging.getLogger(__name__)

# PVE
BOUNTY_FILTER = Q(ref_type__in=BOUNTY_PRIZES)
MISSION_FILTER = Q(ref_type__in=MISSION_REWARD)
ESS_FILTER = Q(ref_type__in=ESS_TRANSFER)
INCURSION_FILTER = Q(ref_type__in=INCURSION)
DAILY_GOAL_REWARD_FILTER = Q(ref_type__in=DAILY_GOAL_REWARD, amount__gt=0)
# COSTS
CONTRACT_COST_FILTER = Q(ref_type__in=CONTRACT, amount__lt=0)
MARKET_COST_FILTER = Q(ref_type__in=MARKET, amount__lt=0)
ASSETS_COST_FILTER = Q(ref_type__in=ASSETS, amount__lt=0)
TRAVELING_COST_FILTER = Q(ref_type__in=TRAVELING, amount__lt=0)
PRODUCTION_COST_FILTER = Q(ref_type__in=PRODUCTION, amount__lt=0)
SKILL_COST_FILTER = Q(ref_type__in=SKILL, amount__lt=0)
INSURANCE_COST_FILTER = Q(ref_type__in=INSURANCE, amount__lt=0)
PLANETARY_COST_FILTER = Q(ref_type__in=PLANETARY, amount__lt=0)
LP_COST_FILTER = Q(ref_type__in=LP, amount__lt=0)
# TRADES
MARKET_INCOME_FILTER = Q(ref_type__in=MARKET, amount__gt=0)
CONTRACT_INCOME_FILTER = Q(ref_type__in=CONTRACT, amount__gt=0)
DONATION_INCOME_FILTER = Q(ref_type__in=DONATION, amount__gt=0)
INSURANCE_INCOME_FILTER = Q(ref_type__in=INSURANCE, amount__gt=0)
MILESTONE_REWARD_FILTER = Q(ref_type__in=MILESTONE_REWARD, amount__gt=0)

PVE_FILTER = BOUNTY_FILTER | ESS_FILTER
MISC_FILTER = (
    MARKET_INCOME_FILTER
    | INSURANCE_INCOME_FILTER
    | MISSION_FILTER
    | INCURSION_FILTER
    | MILESTONE_REWARD_FILTER
)
COST_FILTER = (
    MARKET_COST_FILTER
    | PRODUCTION_COST_FILTER
    | CONTRACT_COST_FILTER
    | LP_COST_FILTER
    | TRAVELING_COST_FILTER
    | ASSETS_COST_FILTER
    | SKILL_COST_FILTER
    | INSURANCE_COST_FILTER
    | PLANETARY_COST_FILTER
)


class CharWalletIncomeFilter(models.QuerySet):
    # PvE - Income
    def annotate_bounty_income(self) -> models.QuerySet:
        return self.annotate(
            bounty_income=Coalesce(
                Sum(
                    "amount",
                    filter=BOUNTY_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_mission_income(self) -> models.QuerySet:
        return self.annotate(
            mission_income=Coalesce(
                Sum(
                    "amount",
                    filter=MISSION_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_incursion_income(self) -> models.QuerySet:
        return self.annotate(
            incursion_income=Coalesce(
                Sum(
                    "amount",
                    filter=INCURSION_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    # Trading - Income
    def annotate_market_income(self) -> models.QuerySet:
        return self.annotate(
            market_income=Coalesce(
                Sum(
                    "amount",
                    filter=MARKET_INCOME_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_contract_income(self) -> models.QuerySet:
        return self.annotate(
            contract_income=Coalesce(
                Sum(
                    "amount",
                    filter=CONTRACT_INCOME_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_donation_income(self, exclude: list = None) -> models.QuerySet:
        if exclude is None:
            exclude = []

        return self.annotate(
            donation_income=Coalesce(
                Sum(
                    "amount",
                    filter=Q(DONATION_INCOME_FILTER) & ~Q(first_party_id__in=exclude),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_insurance_income(self) -> models.QuerySet:
        return self.annotate(
            insurance_income=Coalesce(
                Sum(
                    "amount",
                    filter=INSURANCE_INCOME_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_milestone_income(self) -> models.QuerySet:
        return self.annotate(
            milestone_income=Coalesce(
                Sum(
                    "amount",
                    filter=MILESTONE_REWARD_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )


class CharWalletOutSideFilter(CharWalletIncomeFilter):
    def annotate_miscellaneous(self) -> models.QuerySet:
        return self.annotate(
            miscellaneous=Coalesce(
                Sum(
                    "amount",
                    filter=MISC_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_miscellaneous_with_exclude(self, exclude=None) -> models.QuerySet:
        """Annotate all income together"""
        qs = (
            self.annotate_donation_income(exclude=exclude)
            .annotate_mission_income()
            .annotate_milestone_income()
            .annotate_insurance_income()
            .annotate_market_income()
            .annotate_contract_income()
            .annotate_incursion_income()
        )
        return qs.annotate(
            miscellaneous=Coalesce(
                F("mission_income")
                + F("incursion_income")
                + F("insurance_income")
                + F("market_income")
                + F("contract_income")
                + F("donation_income")
                + F("milestone_income"),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_costs(self) -> models.QuerySet:
        return self.annotate(
            costs=Coalesce(
                Sum(
                    "amount",
                    filter=COST_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )


class CharWalletCostQueryFilter(CharWalletOutSideFilter):
    # Costs
    def annotate_contract_cost(self) -> models.QuerySet:
        return self.annotate(
            contract_cost=Coalesce(
                Sum(
                    "amount",
                    filter=CONTRACT_COST_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_market_cost(self) -> models.QuerySet:
        return self.annotate(
            market_cost=Coalesce(
                Sum(
                    "amount",
                    filter=MARKET_COST_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_asset_cost(self) -> models.QuerySet:
        return self.annotate(
            asset_cost=Coalesce(
                Sum(
                    "amount",
                    filter=ASSETS_COST_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_traveling_cost(self) -> models.QuerySet:
        return self.annotate(
            traveling_cost=Coalesce(
                Sum(
                    "amount",
                    filter=TRAVELING_COST_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_production_cost(self) -> models.QuerySet:
        return self.annotate(
            production_cost=Coalesce(
                Sum(
                    "amount",
                    filter=PRODUCTION_COST_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_skill_cost(self) -> models.QuerySet:
        return self.annotate(
            skill_cost=Coalesce(
                Sum(
                    "amount",
                    filter=SKILL_COST_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_insurance_cost(self) -> models.QuerySet:
        return self.annotate(
            insurance_cost=Coalesce(
                Sum(
                    "amount",
                    filter=INSURANCE_COST_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_planetary_cost(self) -> models.QuerySet:
        return self.annotate(
            planetary_cost=Coalesce(
                Sum(
                    "amount",
                    filter=PLANETARY_COST_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_lp_cost(self) -> models.QuerySet:
        return self.annotate(
            lp_cost=Coalesce(
                Sum(
                    "amount",
                    filter=LP_COST_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )


class CharWalletQuerySet(CharWalletCostQueryFilter):
    pass


class CharWalletManagerBase(models.Manager):
    pass


CharWalletManager = CharWalletManagerBase.from_queryset(CharWalletQuerySet)



import logging
from datetime import datetime

from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q, QuerySet

from allianceauth.eveonline.models import EveAllianceInfo, EveCharacter

from ledger import models

logger = logging.getLogger(__name__)


def get_character(request, character_id) -> tuple[bool, EveCharacter | None]:
    """Get Character and check permissions"""
    perms = True
    if character_id == 0:
        character_id = request.user.profile.main_character.character_id

    try:
        main_char = EveCharacter.objects.get(character_id=character_id)
    except ObjectDoesNotExist:
        return False, None
    except ValueError:
        return None, None

    # check access
    visible = models.CharacterAudit.objects.visible_eve_characters(request.user)
    if main_char not in visible:
        perms = False
    return perms, main_char


def get_corporation(
    request, corporation_id
) -> tuple[bool | None, models.CorporationAudit | None]:
    """Get Corporation and check permissions for each corporation"""
    perms = True

    try:
        main_corp = models.CorporationAudit.objects.get(
            corporation__corporation_id=corporation_id
        )
    except ObjectDoesNotExist:
        return None, None

    # Check access
    visible = models.CorporationAudit.objects.visible_to(request.user)
    if main_corp not in visible:
        perms = False
    return perms, main_corp


def get_alliance(request, alliance_id) -> tuple[bool | None, EveAllianceInfo | None]:
    """Get Alliance and check permissions for each corporation"""
    perms = True

    corporations = models.CorporationAudit.objects.filter(
        corporation__alliance__alliance_id=alliance_id
    )

    if not corporations.exists():
        return None, None

    # Check access
    visible = models.CorporationAudit.objects.visible_to(request.user)

    # Check if there is an intersection between main_corp and visible
    common_corps = corporations.intersection(visible)
    if not common_corps.exists():
        perms = False

    ally = EveAllianceInfo.objects.get(alliance_id=alliance_id)
    return perms, ally


def get_all_corporations_from_alliance(
    request, alliance_id
) -> tuple[bool | None, list[models.CorporationAudit] | None]:
    """Get Alliance and check permissions for each corporation"""
    perms = True

    corporations = models.CorporationAudit.objects.filter(
        corporation__alliance__alliance_id=alliance_id
    )

    if not corporations.exists():
        return None, None

    # Check access
    visible = models.CorporationAudit.objects.visible_to(request.user)

    # Check if there is an intersection between main_corp and visible
    common_corps = corporations.intersection(visible)
    if not common_corps.exists():
        perms = False
    return perms, corporations


def get_alts_queryset(main_char) -> QuerySet[list[EveCharacter]]:
    """Get all alts for a main character."""
    try:
        linked_characters = (
            main_char.character_ownership.user.character_ownerships.all()
        )

        linked_characters = linked_characters.values_list("character_id", flat=True)

        return EveCharacter.objects.filter(id__in=linked_characters)
    except ObjectDoesNotExist:
        return EveCharacter.objects.filter(pk=main_char.pk)


def get_journal_entitys(date: datetime, view, corporations=None) -> set:
    """Get all entity ids from Corporation Journal Queryset filtered by date."""
    filter_date = Q(date__year=date.year)
    if view == "month":
        filter_date &= Q(date__month=date.month)
    elif view == "day":
        filter_date &= Q(date__month=date.month)
        filter_date &= Q(date__day=date.day)

    first_party_ids = models.CorporationWalletJournalEntry.objects.filter(
        filter_date,
        division__corporation__corporation__corporation_id__in=corporations,
    ).values_list("first_party_id", flat=True)

    second_party_ids = models.CorporationWalletJournalEntry.objects.filter(
        filter_date, division__corporation__corporation__corporation_id__in=corporations
    ).values_list("second_party_id", flat=True)

    entity_ids = set(first_party_ids) | set(second_party_ids)

    return entity_ids



"""PvE Views"""

import logging
from datetime import datetime

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import redirect, render

# Django
from django.utils.translation import gettext as _

from allianceauth.eveonline.models import EveCharacter

# Ledger
from ledger.api.helpers import get_character
from ledger.models.characteraudit import CharacterAudit
from ledger.view_helpers.core import add_info_to_context

logger = logging.getLogger(__name__)


@login_required
@permission_required("ledger.basic_access")
def character_ledger_index(request):
    """Character Ledger Index View"""
    context = {}
    context = add_info_to_context(request, context)
    return redirect(
        "ledger:character_ledger", request.user.profile.main_character.character_id
    )


@login_required
@permission_required("ledger.basic_access")
def character_ledger(request, character_id=None):
    """
    Character Ledger
    """
    if character_id is None:
        character_id = request.user.profile.main_character.character_id

    current_year = datetime.now().year
    years = [current_year - i for i in range(6)]

    context = {
        "title": "Character Ledger",
        "years": years,
        "entity_pk": character_id,
        "character_id": character_id,
        "entity_type": "character",
    }
    context = add_info_to_context(request, context)
    return render(request, "ledger/charledger/character_ledger.html", context=context)


@login_required
@permission_required("ledger.basic_access")
def character_overview(request):
    """
    Character Overview
    """

    context = {
        "title": "Character Overview",
    }
    context = add_info_to_context(request, context)
    return render(
        request, "ledger/charledger/admin/character_overview.html", context=context
    )


@login_required
@permission_required("ledger.basic_access")
def character_administration(request, character_id=None):
    """
    Character Administration
    """
    if character_id is None:
        character_id = request.user.profile.main_character.character_id

    perms, character = get_character(request, character_id)

    if not perms:
        msg = _("Permission Denied")
        messages.error(request, msg)
        return redirect("ledger:character_ledger_index")

    linked_characters = character.character_ownership.user.character_ownerships.all()
    linked_characters_ids = linked_characters.values_list(
        "character__character_id", flat=True
    )

    characters = CharacterAudit.objects.filter(
        character__character_id__in=linked_characters_ids
    ).order_by("character__character_name")

    missing_characters = (
        EveCharacter.objects.filter(character_id__in=linked_characters_ids)
        .exclude(
            character_id__in=characters.values_list(
                "character__character_id", flat=True
            )
        )
        .order_by("character_name")
    )

    context = {
        "character_id": character_id,
        "title": "Character Administration",
        "characters": characters,
        "missing_characters": missing_characters,
    }
    context = add_info_to_context(request, context)
    return render(
        request,
        "ledger/charledger/admin/character_administration.html",
        context=context,
    )



import logging

from ninja import NinjaAPI
from ninja.security import django_auth

from django.conf import settings

from ledger.api import ledger

logger = logging.getLogger(__name__)

api = NinjaAPI(
    title="Geuthur API",
    version="0.5.0",
    urls_namespace="ledger:api",
    auth=django_auth,
    csrf=True,
    openapi_url=settings.DEBUG and "/openapi.json" or "",
)

# Add the ledger endpoints
ledger.setup(api)



import logging
from collections import defaultdict
from dataclasses import dataclass

from django.db.models import Q
from django.utils import timezone

from allianceauth.eveonline.models import EveAllianceInfo, EveCharacter

from ledger.models.corporationaudit import (
    CorporationAudit,
)

logger = logging.getLogger(__name__)


@dataclass
# pylint: disable=too-many-instance-attributes
class InformationData:
    """InformationData class to hold the information data."""

    date: timezone.datetime
    view: str
    character: EveCharacter = None
    corporation: CorporationAudit = None
    alliance: EveAllianceInfo = None
    current_date: timezone.datetime = None

    def __post_init__(self):
        self.ledger_date = self.current_date
        self.ledger_date = self.ledger_date.replace(year=self.date.year)
        if self.view == "month":
            self.ledger_date = self.ledger_date.replace(month=self.date.month)
            self.information_date = self.ledger_date.strftime("%B %Y")
        elif self.view == "day":
            self.ledger_date = self.ledger_date.replace(
                month=self.date.month, day=self.date.day
            )
            self.information_date = self.ledger_date.strftime("%d %B %Y")

        if self.view == "year":
            self.current_day = 365
            self.information_date = self.ledger_date.year
        else:
            self.current_day = self.ledger_date.day

        # Information Title
        if self.character is not None:
            self.name = self.character.character_name
            self.id = self.character.character_id
        elif self.corporation is not None:
            self.name = self.corporation.corporation.corporation_name
            self.id = self.corporation.corporation.corporation_id
        elif self.alliance is not None:
            self.name = self.alliance.alliance_name
            self.id = self.alliance.alliance_id
        else:
            self.name = "Unknown"
            self.id = 0

    def get_queryfilter_date(self):
        """Get a query filter date."""
        filter_date = Q(date__year=self.date.year)
        if self.view == "month":
            filter_date &= Q(date__month=self.date.month)
        elif self.view == "day":
            filter_date &= Q(date__month=self.date.month)
            filter_date &= Q(date__day=self.date.day)
        return filter_date

    def _generate_amounts_dict(self, amounts: defaultdict, dict_name: dict):
        """Generate the amounts dictionary."""
        # Convert float values to Decimal before summing
        total_sum = sum(amounts[key]["total_amount"] for key in amounts)

        total_current_day_sum = sum(amounts[key]["total_amount_day"] for key in amounts)

        dict_name.update(
            {
                key: {
                    sub_key: value
                    for sub_key, value in {
                        "total_amount": amounts[key]["total_amount"],
                        "total_amount_day": amounts[key]["total_amount_day"],
                        "average_day": amounts[key]["total_amount"] / self.current_day,
                        "average_hour": (
                            amounts[key]["total_amount"] / self.current_day
                        )
                        / 24,
                        "average_tick": ((amounts[key]["total_amount"]) / 20),
                        "current_day_tick": (amounts[key]["total_amount_day"] / 20),
                        "average_day_tick": (
                            amounts[key]["total_amount"] / self.current_day / 20
                        ),
                        "average_hour_tick": (
                            amounts[key]["total_amount"] / self.current_day
                        )
                        / 24
                        / 20,
                    }.items()
                    if value != 0
                }
                for key in amounts
            }
        )

        dict_name["summary"] = {
            "total_amount": total_sum,
            "total_amount_day": total_sum / self.current_day,
            "total_amount_hour": (total_sum / self.current_day) / 24,
            "total_current_day": total_current_day_sum,
        }
        return dict_name



from .admin import LedgerAdminApiEndpoints
from .ledger import LedgerApiEndpoints
from .planetary import LedgerPlanetaryApiEndpoints


def setup(api):
    LedgerAdminApiEndpoints(api)
    LedgerApiEndpoints(api)
    LedgerPlanetaryApiEndpoints(api)



import logging
from collections import defaultdict
from decimal import Decimal

from django.db.models import Q
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from allianceauth.authentication.models import UserProfile
from allianceauth.eveonline.models import EveCharacter

from ledger.api.api_helper.aggregator import AggregateLedger
from ledger.api.api_helper.billboard_helper import BillboardSystem
from ledger.api.api_helper.information_helper import (
    InformationData,
)
from ledger.api.helpers import get_alts_queryset
from ledger.models.corporationaudit import (
    CorporationAudit,
    CorporationWalletJournalEntry,
)
from ledger.models.general import EveEntity
from ledger.view_helpers.core import events_filter

logger = logging.getLogger(__name__)


class CorporationProcess:
    """JournalProcess class to process the journal entries."""

    def __init__(
        self,
        corporation: CorporationAudit,
        date: timezone.datetime,
        main_character: EveCharacter,
        view=None,
    ):
        self.main_character = main_character
        self.corporation = corporation
        self.date = date
        self.view = view
        self.current_date = timezone.now()

        self._init_journal()

    # pylint: disable=duplicate-code
    def _filter_date(self):
        """Filter the date."""
        filter_date = Q(date__year=self.date.year)
        if self.view == "month":
            filter_date &= Q(date__month=self.date.month)
        elif self.view == "day":
            filter_date &= Q(date__month=self.date.month)
            filter_date &= Q(date__day=self.date.day)
        return filter_date

    def _init_journal(self):
        """Initialize the data for the ledger."""
        self.corporation_journal = CorporationWalletJournalEntry.objects.filter(
            self._filter_date(),
            division__corporation__corporation__corporation_id=self.corporation.corporation.corporation_id,
        )

        # Exclude Corp Tax Events
        self.corporation_journal = events_filter(self.corporation_journal)

        # Get Glances
        self.glance = AggregateLedger(self.corporation_journal)

        # Get Entity IDs from Character or Corporation
        if self.main_character:
            entity_ids = get_alts_queryset(self.main_character)
            self.entity_ids = EveEntity.objects.filter(
                eve_id__in=entity_ids.values_list("character_id", flat=True),
            ).values_list("eve_id", flat=True)
        else:
            self.entity_ids = set(
                self.corporation_journal.values_list("second_party_id", flat=True)
            ) | set(self.corporation_journal.values_list("first_party_id", flat=True))

        self.glance_day = self.corporation_journal.filter(
            date__year=self.current_date.year,
            date__month=self.current_date.month,
            date__day=self.current_date.day,
        )

    def generate_ledger(self):
        accounts = UserProfile.objects.filter(
            main_character__isnull=False,
        ).prefetch_related(
            "user__profile__main_character",
        )

        main_dict = {}
        billboard = BillboardSystem(view=self.view)

        account_totals = defaultdict(int)

        def process_account(account):
            """Helper method to process a single account."""
            alts = account.user.character_ownerships.all().values_list(
                "character__character_id", flat=True
            )
            # Filter out alts that are not in the journal
            alts = set(alts).intersection(self.entity_ids)

            if not alts:
                return None

            bounty = self.glance.aggregate_bounty(alts)
            ess = self.glance.aggregate_ess(alts)
            miscellaneous = (
                self.glance.aggregate_production(
                    alts
                )  # Add Production Amount to Account
                + self.glance.aggregate_miscellaneous(alts)
                + self.glance.aggregate_donation(alts)  # Add Donation Amount to Account
            )

            if bounty > 0 or ess > 0 or miscellaneous > 0:
                data = {
                    "main_id": account.user.profile.main_character.character_id,
                    "main_name": account.user.profile.main_character.character_name,
                    "entity_type": "character",
                    "alt_names": list(alts),
                    "total_amount": bounty,
                    "total_amount_ess": ess,
                    "total_amount_others": miscellaneous,
                    "total_amount_costs": 0,
                }
                account_totals["bounty"] += bounty
                account_totals["ess"] += ess
                account_totals["miscellaneous"] += miscellaneous

                billboard.chord_add_data(
                    data["main_name"], _("Wallet"), (bounty + ess + miscellaneous)
                )
                return data
            return None

        for account in accounts:
            data = process_account(account)
            if data:
                main_dict[data["main_id"]] = data

        # Aggregate totals
        total_bounty = self.glance.aggregate_bounty()
        total_ess = self.glance.aggregate_ess()
        total_miscellaneous = (
            +self.glance.aggregate_miscellaneous()
            + self.glance.aggregate_donation()
            + self.glance.aggregate_corp_withdraw(
                exclude=self.corporation.corporation.corporation_id
            )  # Exclude Intern Transfers
        )
        total_costs = (
            self.glance.aggregate_costs()
            + self.glance.aggregate_corp_withdraw_cost(
                exclude=self.corporation.corporation.corporation_id
            )
        )  # Add Corporation Withdraws Costs

        # Substract account totals from the total amounts
        other_totals = {
            "bounty": total_bounty - account_totals["bounty"],
            "ess": total_ess - account_totals["ess"],
            "miscellaneous": total_miscellaneous - account_totals["miscellaneous"],
            "costs": total_costs,
        }

        if (
            total_bounty > 0
            or total_ess > 0
            or total_miscellaneous > 0
            or total_costs > 0
        ):
            data = {
                "main_id": self.corporation.corporation.corporation_id,
                "main_name": self.corporation.corporation.corporation_name,
                "entity_type": "corporation",
                "alt_names": None,
                "total_amount": other_totals["bounty"],
                "total_amount_ess": other_totals["ess"],
                "total_amount_others": other_totals["miscellaneous"],
                "total_amount_costs": other_totals["costs"],
            }
            main_dict[data["main_id"]] = data
            billboard.chord_add_data(
                data["main_name"],
                _("Wallet"),
                (
                    other_totals["bounty"]
                    + other_totals["ess"]
                    + other_totals["miscellaneous"]
                ),
            )

        # Add Rattingbar for the corporation
        rattingbar_timeline = billboard.create_timeline(self.corporation_journal)
        rattingbar = (
            rattingbar_timeline.annotate_bounty_income()
            .annotate_ess_income()
            .annotate_miscellaneous()
        )
        billboard.create_or_update_results(rattingbar)
        billboard.create_ratting_bar()

        # Order and Handle Overflow
        billboard.chord_handle_overflow()

        output = {
            "ratting": sorted(list(main_dict.values()), key=lambda x: x["main_name"]),
            "billboard": billboard.dict,
            "total": {
                "total_amount": total_bounty,
                "total_amount_ess": total_ess,
                "total_amount_others": total_miscellaneous,
                "total_amount_costs": total_costs,
                "total_amount_all": total_bounty
                + total_ess
                + total_miscellaneous
                + total_costs,
            },
        }

        return output

    def generate_template(self):
        """Generate the information for the corporation"""
        information_dict = {}
        exclude = None

        # Create the Ledger
        ledger_data = InformationData(
            character=self.main_character,
            corporation=self.corporation,
            date=self.date,
            view=self.view,
            current_date=self.current_date,
        )

        if self.corporation is not None:
            exclude = self.corporation.corporation.corporation_id

        day_aggregate = AggregateLedger(self.glance_day)

        amounts = defaultdict(lambda: defaultdict(Decimal))

        amounts["bounty_income"] = {
            "total_amount": self.glance.aggregate_bounty(self.entity_ids),
            "total_amount_day": day_aggregate.aggregate_bounty(self.entity_ids),
        }

        amounts["ess_income"] = {
            "total_amount": self.glance.aggregate_ess(self.entity_ids),
            "total_amount_day": day_aggregate.aggregate_ess(self.entity_ids),
        }

        amounts["mission_income"] = {
            "total_amount": self.glance.aggregate_mission(self.entity_ids),
            "total_amount_day": day_aggregate.aggregate_mission(self.entity_ids),
        }

        amounts["market_income"] = {
            "total_amount": self.glance.aggregate_market(second_party=self.entity_ids),
            "total_amount_day": day_aggregate.aggregate_market(
                second_party=self.entity_ids
            ),
        }

        amounts["contract_income"] = {
            "total_amount": self.glance.aggregate_contract(self.entity_ids),
            "total_amount_day": day_aggregate.aggregate_contract(self.entity_ids),
        }

        amounts["donation_income"] = {
            "total_amount": self.glance.aggregate_donation(self.entity_ids)
            + self.glance.aggregate_corp_withdraw(self.entity_ids, exclude=exclude),
            "total_amount_day": day_aggregate.aggregate_donation(self.entity_ids)
            + day_aggregate.aggregate_corp_withdraw(self.entity_ids, exclude=exclude),
        }

        amounts["insurance_income"] = {
            "total_amount": self.glance.aggregate_insurance(self.entity_ids),
            "total_amount_day": day_aggregate.aggregate_insurance(self.entity_ids),
        }

        amounts["daily_goal_income"] = {
            "total_amount": self.glance.aggregate_daily_goal(self.entity_ids),
            "total_amount_day": day_aggregate.aggregate_daily_goal(self.entity_ids),
        }

        amounts["citadel_income"] = {
            "total_amount": (
                self.glance.aggregate_production(self.entity_ids)
                + self.glance.aggregate_traveling(self.entity_ids)
            ),
            "total_amount_day": (
                day_aggregate.aggregate_production(self.entity_ids)
                + day_aggregate.aggregate_traveling(self.entity_ids)
            ),
        }

        # Only Corporation will have costs
        if not self.main_character:
            amounts["market_cost"] = {
                "total_amount": self.glance.aggregate_market_cost(self.entity_ids),
                "total_amount_day": day_aggregate.aggregate_market_cost(
                    self.entity_ids
                ),
            }

            amounts["rental_cost"] = {
                "total_amount": self.glance.aggregate_rental(self.entity_ids),
                "total_amount_day": day_aggregate.aggregate_rental(self.entity_ids),
            }

            amounts["donation_cost"] = {
                "total_amount": self.glance.aggregate_corp_withdraw_cost(
                    self.entity_ids, exclude=exclude
                ),
                "total_amount_day": day_aggregate.aggregate_corp_withdraw_cost(
                    self.entity_ids, exclude=exclude
                ),
            }

            amounts["contract_cost"] = {
                "total_amount": self.glance.aggregate_contract_cost(self.entity_ids),
                "total_amount_day": day_aggregate.aggregate_contract_cost(
                    self.entity_ids
                ),
            }

            amounts["production_cost"] = {
                "total_amount": self.glance.aggregate_production_cost(self.entity_ids),
                "total_amount_day": day_aggregate.aggregate_production_cost(
                    self.entity_ids
                ),
            }

        information_dict.update(
            {
                "main_id": ledger_data.id,
                "main_name": ledger_data.name,
                "date": ledger_data.information_date,
            }
        )
        information_dict = ledger_data._generate_amounts_dict(amounts, information_dict)
        return information_dict



"""App URLs"""

from django.urls import path, re_path

from ledger.api import api
from ledger.views.alliance import alliance_ledger
from ledger.views.alliance.add_ally import add_ally
from ledger.views.character import character_ledger, planetary
from ledger.views.character.add_char import add_char
from ledger.views.corporation import corp_events, corporation_ledger
from ledger.views.corporation.add_corp import add_corp

# AA Example App
from ledger.views.index import index

app_name: str = "ledger"

urlpatterns = [
    path("", index, name="index"),
    # -- Character Audit
    path("char/add/", add_char, name="add_char"),
    # -- Corporation Audit
    path("corporation/add/", add_corp, name="add_corp"),
    # -- -- Alliance Ledger
    path(
        "alliance_ledger/",
        alliance_ledger.alliance_ledger_index,
        name="alliance_ledger_index",
    ),
    path(
        "alliance_ledger/<int:alliance_id>/",
        alliance_ledger.alliance_ledger,
        name="alliance_ledger",
    ),
    path(
        "alliance_ledger/<int:alliance_id>/view/administration/",
        alliance_ledger.alliance_administration,
        name="alliance_administration",
    ),
    path(
        "alliance_overview/",
        alliance_ledger.alliance_overview,
        name="alliance_overview",
    ),
    path("alliance/add/", add_ally, name="add_ally"),
    # -- -- Corporation Ledger
    path(
        "corporation_ledger/",
        corporation_ledger.corporation_ledger_index,
        name="corporation_ledger_index",
    ),
    path(
        "corporation_ledger/<int:corporation_id>/",
        corporation_ledger.corporation_ledger,
        name="corporation_ledger",
    ),
    path(
        "corporation_ledger/<int:corporation_id>/view/administration/",
        corporation_ledger.corporation_administration,
        name="corporation_administration",
    ),
    path(
        "corporation_overview/",
        corporation_ledger.corporation_overview,
        name="corporation_overview",
    ),
    # -- -- Character Ledger
    path(
        "character_ledger/",
        character_ledger.character_ledger_index,
        name="character_ledger_index",
    ),
    path(
        "character_ledger/<int:character_id>/",
        character_ledger.character_ledger,
        name="character_ledger",
    ),
    path(
        "character_ledger/<int:character_id>/view/administration/",
        character_ledger.character_administration,
        name="character_administration",
    ),
    path(
        "character_overview/",
        character_ledger.character_overview,
        name="character_overview",
    ),
    # -- -- Events
    path("events/", corp_events.events_index, name="events_index"),
    path("events/admin/", corp_events.events_admin, name="event_admin"),
    path("events/create/", corp_events.create_event, name="create_event"),
    path("events/<int:event_id>/edit/", corp_events.edit_event, name="edit_event"),
    path(
        "events/<int:event_id>/delete/", corp_events.delete_event, name="delete_event"
    ),
    path("events/ajax/load_events", corp_events.load_events, name="load_events"),
    # -- -- Planetary
    path(
        "planetary_ledger/",
        planetary.planetary_ledger_index,
        name="planetary_ledger_index",
    ),
    path(
        "planetary_ledger/<int:character_id>/",
        planetary.planetary_ledger,
        name="planetary_ledger",
    ),
    path(
        "planetary/switch_alarm/",
        planetary.switch_alarm,
        name="switch_alarm",
    ),
    path(
        "planetary_overview/", planetary.planetary_overview, name="planetary_overview"
    ),
    # -- API System
    re_path(r"^api/", api.urls),
]



from django.template import Context, Template
from django.test import TestCase, override_settings

from ledger import __version__
from ledger.view_helpers.static_files import calculate_integrity_hash


class TestVersionedStatic(TestCase):
    @override_settings(DEBUG=False)
    def test_versioned_static(self):
        context = Context(dict_={"version": __version__})
        template_to_render = Template(
            template_string=(
                "{% load ledger %}"
                "{% ledger_static 'css/ledger.css' %}"
                "{% ledger_static 'js/character.js' %}"
            )
        )

        rendered_template = template_to_render.render(context=context)

        expected_static_css_src = (
            f'/static/ledger/css/ledger.css?v={context["version"]}'
        )
        expected_static_css_src_integrity = calculate_integrity_hash("css/ledger.css")
        expected_static_js_src = (
            f'/static/ledger/js/character.js?v={context["version"]}'
        )
        expected_static_js_src_integrity = calculate_integrity_hash("js/character.js")

        self.assertIn(member=expected_static_css_src, container=rendered_template)
        self.assertIn(
            member=expected_static_css_src_integrity, container=rendered_template
        )
        self.assertIn(member=expected_static_js_src, container=rendered_template)
        self.assertIn(
            member=expected_static_js_src_integrity, container=rendered_template
        )

    @override_settings(DEBUG=True)
    def test_versioned_static_with_debug_enabled(self) -> None:
        context = Context({"version": __version__})
        template_to_render = Template(
            template_string=("{% load ledger %}" "{% ledger_static 'css/ledger.css' %}")
        )

        rendered_template = template_to_render.render(context=context)

        expected_static_css_src = (
            f'/static/ledger/css/ledger.css?v={context["version"]}'
        )

        self.assertIn(member=expected_static_css_src, container=rendered_template)
        self.assertNotIn(member="integrity=", container=rendered_template)

    @override_settings(DEBUG=False)
    def test_invalid_file_type(self) -> None:
        context = Context({"version": __version__})
        template_to_render = Template(
            template_string=(
                "{% load ledger %}" "{% ledger_static 'invalid/invalid.txt' %}"
            )
        )

        with self.assertRaises(ValueError):
            template_to_render.render(context=context)



"""
Decorators
"""

import logging
import time
from functools import wraps

from app_utils.esi import EsiDailyDowntime, fetch_esi_status

from ledger.app_settings import IS_TESTING

logger = logging.getLogger(__name__)


def when_esi_is_available(func):
    """Make sure the decorated task only runs when esi is available.

    Raise exception when ESI is offline.
    Complete the task without running it when downtime is detected.

    Automatically disabled during tests.
    """

    @wraps(func)
    def outer(*args, **kwargs):
        if IS_TESTING is not True:
            try:
                fetch_esi_status().raise_for_status()
            except EsiDailyDowntime:
                logger.info("Daily Downtime detected. Aborting.")
                return None  # function will not run

        return func(*args, **kwargs)

    return outer


def log_timing(logs):
    """
    Ein Dekorator, der die Ausführungszeit einer Funktion misst und in die Logdatei schreibt.
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            start_time = time.time()
            result = func(*args, **kwargs)
            end_time = time.time()
            logs.debug(
                "TIME: %s run for %s seconds with args: %s",
                end_time - start_time,
                func.__name__,
                args,
            )
            return result

        return wrapper

    return decorator



"""
Character Helpers
"""

import logging
from datetime import timedelta

from django.utils import timezone
from eveuniverse.models import EveType

from ledger.decorators import log_timing
from ledger.errors import TokenError
from ledger.models.characteraudit import (
    CharacterAudit,
    CharacterMiningLedger,
    CharacterWalletJournalEntry,
)
from ledger.models.general import EveEntity
from ledger.providers import esi
from ledger.task_helpers.core_helpers import get_token
from ledger.task_helpers.etag_helpers import (
    HTTPGatewayTimeoutError,
    NotModifiedError,
    etag_results,
)

logger = logging.getLogger(__name__)


# pylint: disable=too-many-locals
@log_timing(logger)
def update_character_wallet(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(character__character_id=character_id)
    logger.debug(
        "Updating wallet transactions for: %s", audit_char.character.character_name
    )

    req_scopes = ["esi-wallet.read_character_wallet.v1"]

    token = get_token(character_id, req_scopes)

    if not token:
        logger.info(
            "No Wallet Token for %s",
            audit_char.character.character_name,
        )
        audit_char.is_active()
        return "No Tokens"

    try:
        journal_items_ob = esi.client.Wallet.get_characters_character_id_wallet_journal(
            character_id=character_id
        )

        journal_items = etag_results(
            journal_items_ob, token, force_refresh=force_refresh
        )

        _current_journal = CharacterWalletJournalEntry.objects.filter(
            character=audit_char
        ).values_list(
            "entry_id", flat=True
        )  # TODO add time filter
        _current_eve_ids = list(
            EveEntity.objects.all().values_list("eve_id", flat=True)
        )

        _new_names = []

        items = []
        for item in journal_items:
            if item.get("id") not in _current_journal:
                if item.get("second_party_id") not in _current_eve_ids:
                    _new_names.append(item.get("second_party_id"))
                    _current_eve_ids.append(item.get("second_party_id"))
                if item.get("first_party_id") not in _current_eve_ids:
                    _new_names.append(item.get("first_party_id"))
                    _current_eve_ids.append(item.get("first_party_id"))

                # pylint: disable=duplicate-code
                asset_item = CharacterWalletJournalEntry(
                    character=audit_char,
                    amount=item.get("amount"),
                    balance=item.get("balance"),
                    context_id=item.get("context_id"),
                    context_id_type=item.get("context_id_type"),
                    date=item.get("date"),
                    description=item.get("description"),
                    first_party_id=item.get("first_party_id"),
                    entry_id=item.get("id"),
                    reason=item.get("reason"),
                    ref_type=item.get("ref_type"),
                    second_party_id=item.get("second_party_id"),
                    tax=item.get("tax"),
                    tax_receiver_id=item.get("tax_receiver_id"),
                )
                items.append(asset_item)

        created_names = EveEntity.objects.create_bulk_from_esi(_new_names)

        wallet_ballance = esi.client.Wallet.get_characters_character_id_wallet(
            character_id=character_id, token=token.valid_access_token()
        ).result()

        audit_char.balance = wallet_ballance
        audit_char.save()

        if created_names:
            CharacterWalletJournalEntry.objects.bulk_create(items)
        else:
            raise TokenError("ESI Fail")

        logger.debug(
            "Finished wallet transactions for: %s", audit_char.character.character_name
        )
    except NotModifiedError:
        logger.debug("No New wallet data for: %s", audit_char.character.character_name)
    except HTTPGatewayTimeoutError:
        logger.debug("Gateway Timeout for: %s", audit_char.character.character_name)

    audit_char.last_update_wallet = timezone.now()
    audit_char.save()
    audit_char.is_active()
    return "Success"


@log_timing(logger)
def update_character_mining(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(character__character_id=character_id)
    logger.debug("Updating Mining for: %s", audit_char.character.character_name)

    req_scopes = ["esi-industry.read_character_mining.v1"]

    token = get_token(character_id, req_scopes)

    if not token:
        logger.info(
            "No Mining Token for %s, Deactivate Character",
            audit_char.character.character_name,
        )
        audit_char.is_active()
        return "No Tokens"
    try:
        mining_op = esi.client.Industry.get_characters_character_id_mining(
            character_id=character_id
        )

        ledger = etag_results(mining_op, token, force_refresh=force_refresh)

        existings_pks = set(
            CharacterMiningLedger.objects.filter(
                character=audit_char, date__gte=timezone.now() - timedelta(days=30)
            ).values_list("id", flat=True)
        )
        type_ids = set()
        new_events = []
        old_events = []
        for event in ledger:

            type_ids.add(event.get("type_id"))
            pk = CharacterMiningLedger.create_primary_key(character_id, event)
            _e = CharacterMiningLedger(
                character=audit_char,
                id=pk,
                date=event.get("date"),
                type_id=event.get("type_id"),
                system_id=event.get("solar_system_id"),
                quantity=event.get("quantity"),
            )
            if pk in existings_pks:
                old_events.append(_e)
            else:
                new_events.append(_e)

        EveType.objects.bulk_get_or_create_esi(ids=list(type_ids))

        if new_events:
            CharacterMiningLedger.objects.bulk_create(new_events, ignore_conflicts=True)

        if old_events:
            CharacterMiningLedger.objects.bulk_update(old_events, fields=["quantity"])
        logger.debug("Finished Mining for: %s", audit_char.character.character_name)
    except NotModifiedError:
        logger.debug("No New Mining for: %s", audit_char.character.character_name)
    except HTTPGatewayTimeoutError:
        logger.debug("Gateway Timeout for: %s", audit_char.character.character_name)
        return "Gateway Timeout"

    audit_char.last_update_mining = timezone.now()
    audit_char.save()
    audit_char.is_active()
    return "Success"



import logging
from decimal import Decimal

from django.db import models
from django.db.models import DecimalField, F, Q, Sum, Value
from django.db.models.functions import Coalesce, Round

from ledger import app_settings
from ledger.constants import (
    ASSETS,
    BOUNTY_PRIZES,
    CONTRACT,
    DAILY_GOAL_REWARD,
    DONATION,
    ESS_TRANSFER,
    INCURSION,
    INSURANCE,
    LP,
    MARKET,
    MISSION_REWARD,
    PRODUCTION,
    RENTAL,
    SKILL,
    TRAVELING,
)

logger = logging.getLogger(__name__)

# Filters
BOUNTY_FILTER = Q(ref_type__in=BOUNTY_PRIZES, amount__gt=0)
ESS_FILTER = Q(ref_type__in=ESS_TRANSFER, amount__gt=0)
INCURSION_FILTER = Q(ref_type__in=INCURSION, amount__gt=0)
MISSION_FILTER = Q(ref_type__in=MISSION_REWARD, amount__gt=0)
DAILY_GOAL_REWARD_FILTER = Q(ref_type__in=DAILY_GOAL_REWARD, amount__gt=0)
CITADEL_FILTER = Q(ref_type__in=PRODUCTION, amount__gt=0)

MISC_FILTER = Q(
    ref_type__in=[
        *ASSETS,
        *CONTRACT,
        *DAILY_GOAL_REWARD,
        *DONATION,
        *INCURSION,
        *INSURANCE,
        *MISSION_REWARD,
        *PRODUCTION,
        *MARKET,
        *TRAVELING,
        *LP,
    ],
    amount__gt=0,
)

COSTS_FILTER = Q(
    ref_type__in=[
        *ASSETS,
        *CONTRACT,
        *INSURANCE,
        *LP,
        *MARKET,
        *TRAVELING,
        *PRODUCTION,
        *SKILL,
        *RENTAL,
    ],
    amount__lt=0,
)


class CorpWalletQuerySet(models.QuerySet):
    def _convert_corp_tax(self, ess: models.QuerySet) -> Decimal:
        """Convert corp tax to correct amount for character ledger"""
        amount = (ess / app_settings.LEDGER_CORP_TAX) * (
            100 - app_settings.LEDGER_CORP_TAX
        )
        return amount

    def annotate_bounty_income(self) -> models.QuerySet:
        return self.annotate(
            bounty_income=Coalesce(
                Sum(
                    "amount",
                    filter=(BOUNTY_FILTER),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_ess_income(self, is_character: bool = False) -> models.QuerySet:
        if is_character:
            return self.annotate(
                ess_income=Round(
                    Coalesce(
                        Sum(
                            self._convert_corp_tax(F("amount")),
                            filter=(ESS_FILTER),
                        ),
                        Value(0),
                        output_field=DecimalField(),
                    ),
                    precision=2,
                )
            )
        return self.annotate(
            ess_income=Coalesce(
                Sum(
                    F("amount"),
                    filter=(ESS_FILTER),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    # pylint: disable=duplicate-code
    def annotate_miscellaneous(self) -> models.QuerySet:
        return self.annotate(
            miscellaneous=Coalesce(
                Sum(
                    "amount",
                    filter=(MISC_FILTER),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_costs(self) -> models.QuerySet:
        return self.annotate(
            costs=Coalesce(
                Sum(
                    "amount",
                    filter=Q(COSTS_FILTER) & Q(amount__lt=0),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )


class CorpWalletManagerBase(models.Manager):
    pass


CorpWalletManager = CorpWalletManagerBase.from_queryset(CorpWalletQuerySet)



from unittest.mock import MagicMock, patch

from django.core.cache import cache
from django.test import TestCase

from ledger.view_helpers.core import (
    _storage_key,
    add_info_to_context,
    calculate_ess_stolen,
    calculate_ess_stolen_amount,
    delete_cache,
    events_filter,
    get_cache_stale,
    set_cache,
)

MODULE_PATH = "ledger.view_helpers.core"


class TestViewHelpers(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    def test_add_info_to_context(self):
        # given
        request = MagicMock()
        request.user = MagicMock()
        request.user.id = 99999999

        context = {"theme": None}
        # when
        result = add_info_to_context(request, context)
        # then
        self.assertEqual(result, context)

    def test_get_cache_stale(self):
        result = get_cache_stale("test")
        self.assertFalse(result)

    @patch.object(cache, "get", return_value="test_value")
    def test_get_cache_stale_with_data(self, mock_cache_get):
        # given
        key_name = "test_key"
        # when
        result = get_cache_stale(key_name)
        # then
        mock_cache_get.assert_called_once_with(key=_storage_key(key_name))
        self.assertEqual(result, "test_value")

    def test_set_cache(self):
        result = set_cache("test", "test", 1)
        self.assertIsNone(result)

    def test_delete_cache(self):
        result = delete_cache("test")
        self.assertIsNone(result)

    def test_storage_key(self):
        result = _storage_key("test")
        self.assertEqual(result, "ledger_storage_test")

    def test_calculate_ess_stolen_amount(self):
        result = calculate_ess_stolen_amount(1000, 1000)
        self.assertEqual(result, 0)

    def test_calculate_ess_stolen_amount_ess_different(self):
        result = calculate_ess_stolen_amount(1000, 200)
        self.assertEqual(result, 93)

    def test_calculate_ess_stolen_amount_with_exception(self):
        result = calculate_ess_stolen_amount("test", "test")
        self.assertEqual(result, 0)

    def test_calculate_ess_stolen(self):
        result = calculate_ess_stolen(
            {
                "bounty": {"total_amount": 1000},
                "ess": {"total_amount": 1000},
                "stolen": {},
            }
        )

        excepted_result = {
            "bounty": {"total_amount": 1000},
            "ess": {"total_amount": 1000},
            "stolen": {"total_amount": 0},
        }

        self.assertEqual(result, excepted_result)

    def test_calculate_ess_stolen_with_exception(self):
        result = calculate_ess_stolen({"bounty": {"total_amount": 1000}})
        self.assertEqual(result, {"bounty": {"total_amount": 1000}})



from unittest.mock import patch

from django.test import TestCase
from django.utils import timezone

from ledger.models.corporationaudit import CorporationWalletJournalEntry
from ledger.models.events import Events
from ledger.models.general import EveEntity
from ledger.tests.testdata.generate_corporationaudit import (
    add_corporationaudit_corporation_to_user,
    create_user_from_evecharacter,
)
from ledger.tests.testdata.generate_events import create_event_1_day
from ledger.tests.testdata.generate_walletjournal import (
    create_division,
    create_wallet_journal_entry,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_eveentity import load_eveentity
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse
from ledger.view_helpers.core import events_filter

MODULE_PATH = "ledger.view_helpers.core"


class TestViewHelpersEvents(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()
        load_eveentity()

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001, permissions=["ledger.basic_access"]
        )
        cls.audit = add_corporationaudit_corporation_to_user(
            cls.user, cls.character_ownership.character.character_id
        )
        cls.division = create_division(
            corporation=cls.audit, name="MEGA KONTO", balance=1000000, division=1
        )
        cls.eve_character_first_party = EveEntity.objects.get(eve_id=1001)
        cls.eve_character_second_party = EveEntity.objects.get(eve_id=1002)
        create_wallet_journal_entry(
            journal_type="corporation",
            division=cls.division,
            entry_id=1,
            amount=1000,
            balance=1000000,
            date=timezone.datetime.replace(
                timezone.now(),
                year=2024,
                month=1,
                day=1,
                hour=0,
                minute=0,
                second=0,
                microsecond=0,
            ),
            description="Test",
            first_party=cls.eve_character_first_party,
            second_party=cls.eve_character_second_party,
            ref_type="ess_escrow_transfer",
        )
        create_wallet_journal_entry(
            journal_type="corporation",
            division=cls.division,
            entry_id=2,
            amount=5000,
            balance=1000000,
            date=timezone.datetime.replace(
                timezone.now(),
                year=2024,
                month=1,
                day=2,
                hour=10,
                minute=0,
                second=0,
                microsecond=0,
            ),
            description="Is Filtered from Events",
            first_party=cls.eve_character_first_party,
            second_party=cls.eve_character_second_party,
            ref_type="ess_escrow_transfer",
        )
        create_wallet_journal_entry(
            journal_type="corporation",
            division=cls.division,
            entry_id=2,
            amount=5000,
            balance=1000000,
            date=timezone.datetime.replace(
                timezone.now(),
                year=2024,
                month=1,
                day=5,
                hour=10,
                minute=0,
                second=0,
                microsecond=0,
            ),
            description="Is Filtered from Events",
            first_party=cls.eve_character_first_party,
            second_party=cls.eve_character_second_party,
            ref_type="ess_escrow_transfer",
        )

    def test_events_filter_should_filter(self):
        # given
        create_event_1_day(
            char_ledger=True,
            date_start=timezone.datetime.replace(
                timezone.now(),
                year=2024,
                month=1,
                day=2,
                hour=0,
                minute=0,
                second=0,
                microsecond=0,
            ),
        )
        journal = CorporationWalletJournalEntry.objects.all()
        # when
        journal_result = events_filter(journal)
        # then
        self.assertEqual(journal_result.count(), 2)
        self.assertEqual(journal_result[0].entry_id, 1)

    def test_events_filter_should_not_filter(self):
        # given
        create_event_1_day(
            char_ledger=False,
            date_start=timezone.datetime.replace(
                timezone.now(),
                year=2024,
                month=1,
                day=2,
                hour=0,
                minute=0,
                second=0,
                microsecond=0,
            ),
        )
        journal = CorporationWalletJournalEntry.objects.all()
        # when
        journal_result = events_filter(journal)
        # then
        self.assertEqual(journal_result.count(), 3)

    def test_events_filter_should_filter_all_events(self):
        # given
        create_event_1_day(
            char_ledger=True,
            date_start=timezone.datetime.replace(
                timezone.now(),
                year=2024,
                month=1,
                day=2,
                hour=0,
                minute=0,
                second=0,
                microsecond=0,
            ),
        )
        create_event_1_day(
            char_ledger=True,
            date_start=timezone.datetime.replace(
                timezone.now(),
                year=2024,
                month=1,
                day=5,
                hour=0,
                minute=0,
                second=0,
                microsecond=0,
            ),
        )
        journal = CorporationWalletJournalEntry.objects.all()
        # when
        journal_result = events_filter(journal)
        # then
        self.assertEqual(journal_result.count(), 1)



import logging
from decimal import Decimal

from django.db.models import Q, QuerySet, Sum

from ledger import app_settings
from ledger.constants import (
    ASSETS,
    BOUNTY_PRIZES,
    CONTRACT,
    CORP_WITHDRAW,
    DAILY_GOAL_REWARD,
    DONATION,
    ESS_TRANSFER,
    INCURSION,
    INSURANCE,
    LP,
    MARKET,
    MILESTONE_REWARD,
    MISSION_REWARD,
    PLANETARY,
    PRODUCTION,
    RENTAL,
    SKILL,
    TRAVELING,
)
from ledger.models.characteraudit import CharacterMiningLedger
from ledger.models.corporationaudit import (
    CorporationWalletJournalEntry,
)

logger = logging.getLogger(__name__)


def convert_corp_tax(amount: Decimal) -> Decimal:
    """Convert corp tax to correct amount for character ledger"""
    return (amount / app_settings.LEDGER_CORP_TAX) * (
        100 - app_settings.LEDGER_CORP_TAX
    )


class AggregateCore:
    def __init__(self, qs: QuerySet):
        self.queryset = qs


class AggegratePvE(AggregateCore):
    """Aggregate PvE class to process PvE data."""

    def aggregate_bounty(self, second_party=None):
        """Aggregate bounty data."""
        qs = self.queryset

        if isinstance(second_party, int):
            second_party = [second_party]

        if second_party:
            qs = qs.filter(second_party__in=second_party)

        return (
            qs.filter(Q(ref_type__in=BOUNTY_PRIZES) & Q(amount__gt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )

    def aggregate_ess(self, second_party=None, is_character=False):
        """Aggregate ESS data.
        This method is only for CorporationWalletJournalEntry.
        """
        qs = self.queryset

        if isinstance(second_party, int):
            second_party = [second_party]

        if second_party:
            qs = qs.filter(second_party__in=second_party)

        ess_amount = (
            qs.filter(Q(ref_type__in=ESS_TRANSFER) & Q(amount__gt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )

        if is_character:
            # Convert the ess amount to the correct value
            ess_amount = int(convert_corp_tax(ess_amount))

        return ess_amount

    def aggregate_mission(self, second_party=None):
        """Aggregate mission data."""
        qs = self.queryset

        if isinstance(second_party, int):
            second_party = [second_party]

        if second_party:
            qs = qs.filter(second_party__in=second_party)

        return (
            qs.filter(Q(ref_type__in=MISSION_REWARD) & Q(amount__gt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )

    def aggregate_incursion(self, second_party=None):
        """Aggregate incursion data."""
        qs = self.queryset

        if isinstance(second_party, int):
            second_party = [second_party]

        if second_party:
            qs = qs.filter(second_party__in=second_party)

        return (
            qs.filter(Q(ref_type__in=INCURSION) & Q(amount__gt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )


class AggregateMining(AggregateCore):
    """Aggregate Mining class to process mining data."""

    def aggregate_mining(self, characters=None):
        """Aggregate mining data."""
        if self.queryset.model != CharacterMiningLedger:
            raise TypeError("This method is only for CharacterMiningLedger")
        qs = self.queryset

        if isinstance(characters, int):
            characters = [characters]

        if characters:
            qs = qs.filter(character__character__character_id__in=characters)

        qs = qs.annotate_mining()

        if qs:
            return int(qs[0]["total_amount"])
        return 0


class AggregateCosts(AggregateCore):

    def aggregate_market_cost(self, first_party=None):
        """Aggregate market cost data."""
        qs = self.queryset

        if isinstance(first_party, int):
            first_party = [first_party]

        if first_party:
            # Filter Costs
            qs = qs.filter(first_party__in=first_party)

        return (
            qs.filter(Q(ref_type__in=MARKET) & Q(amount__lt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )

    def aggregate_insurance_cost(self, first_party=None):
        """Aggregate insurance cost data."""
        qs = self.queryset

        if isinstance(first_party, int):
            first_party = [first_party]

        if first_party:
            # Filter Costs
            qs = qs.filter(first_party__in=first_party)

        return (
            qs.filter(Q(ref_type__in=INSURANCE) & Q(amount__lt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )

    def aggregate_contract_cost(self, first_party=None):
        """Aggregate contract data."""
        qs = self.queryset

        if isinstance(first_party, int):
            first_party = [first_party]

        if first_party:
            # Filter Costs
            qs = qs.filter(first_party__in=first_party)

        return (
            qs.filter(Q(ref_type__in=CONTRACT) & Q(amount__lt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )

    def aggregate_lp_cost(self, first_party=None):
        """Aggregate LP cost data."""
        qs = self.queryset

        if isinstance(first_party, int):
            first_party = [first_party]

        if first_party:
            # Filter Costs
            qs = qs.filter(first_party__in=first_party)

        return (
            qs.filter(Q(ref_type__in=LP) & Q(amount__lt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )

    def aggregate_production_cost(self, first_party=None):
        """Aggregate production cost data."""
        qs = self.queryset

        if isinstance(first_party, int):
            first_party = [first_party]

        if first_party:
            # Filter Costs
            qs = qs.filter(first_party__in=first_party)

        return (
            qs.filter(Q(ref_type__in=PRODUCTION) & Q(amount__lt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )

    def aggregate_costs(self, first_party=None):
        """Aggregate costs data."""
        qs = self.queryset

        if isinstance(first_party, int):
            first_party = [first_party]

        if first_party:
            # Filter Costs
            qs = qs.filter(first_party__in=first_party)

        return (
            qs.filter(
                Q(
                    ref_type__in=[
                        *MARKET,
                        *CONTRACT,
                        *TRAVELING,
                        *PRODUCTION,
                        *ASSETS,
                        *SKILL,
                        *PLANETARY,
                        *LP,
                        *INSURANCE,
                        *RENTAL,
                    ]
                )
                & Q(amount__lt=0)
            ).aggregate(total=Sum("amount"))["total"]
            or 0
        )


class AggregateCorporation(AggregateCore):
    """Aggregate Corporation class to process corporation data."""

    def aggregate_daily_goal(self, second_party=None, is_character=False):
        """
        Aggregate daily goal data.
        This method is only for CorporationWalletJournalEntry.
        """
        if self.queryset.model != CorporationWalletJournalEntry:
            raise TypeError("This method is only for CorporationWalletJournalEntry")

        qs = self.queryset

        if isinstance(second_party, int):
            second_party = [second_party]

        if second_party:
            qs = qs.filter(second_party__in=second_party)

        daily_goal = (
            qs.filter(Q(ref_type__in=DAILY_GOAL_REWARD, amount__gt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )

        if is_character:
            # Convert the ess amount to the correct value
            daily_goal = int(convert_corp_tax(daily_goal))

        return daily_goal

    def aggregate_corp_withdraw(self, first_party=None, exclude=None):
        """Aggregate corp withdraw cost data."""
        qs = self.queryset

        if isinstance(first_party, int):
            first_party = [first_party]

        if first_party:
            # Filter Costs
            qs = qs.filter(first_party__in=first_party)

        if isinstance(exclude, int):
            exclude = [exclude]

        if exclude:
            qs = qs.exclude(second_party_id__in=exclude)

        return (
            qs.filter(Q(ref_type__in=CORP_WITHDRAW) & Q(amount__gt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )

    def aggregate_corp_withdraw_cost(self, first_party=None, exclude=None):
        """Aggregate corp withdraw cost data."""
        qs = self.queryset

        if isinstance(first_party, int):
            first_party = [first_party]

        if first_party:
            # Filter Costs
            qs = qs.filter(first_party__in=first_party)

        if isinstance(exclude, int):
            exclude = [exclude]

        if exclude:
            qs = qs.exclude(second_party_id__in=exclude)

        return (
            qs.filter(Q(ref_type__in=CORP_WITHDRAW) & Q(amount__lt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )


class AggregateLedger(
    AggegratePvE, AggregateCosts, AggregateCorporation, AggregateCore
):
    """Aggregate Ledger class to process ledger data."""

    def aggregate_insurance(self, second_party=None):
        """Aggregate insurance data."""
        qs = self.queryset

        if isinstance(second_party, int):
            second_party = [second_party]

        if second_party:
            # Filter Insurance
            qs = qs.filter(second_party__in=second_party)

        return (
            qs.filter(Q(ref_type__in=INSURANCE) & Q(amount__gt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )

    def aggregate_market(self, second_party=None):
        """Aggregate market data."""
        qs = self.queryset

        if isinstance(second_party, int):
            second_party = [second_party]

        if second_party:
            # Filter Market
            qs = qs.filter(second_party__in=second_party)

        return (
            qs.filter(Q(ref_type__in=MARKET) & Q(amount__gt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )

    def aggregate_contract(self, second_party=None):
        """Aggregate contract data."""
        qs = self.queryset

        if isinstance(second_party, int):
            second_party = [second_party]

        if second_party:
            # Filter Contract
            qs = qs.filter(second_party__in=second_party)

        return (
            qs.filter(Q(ref_type__in=CONTRACT) & Q(amount__gt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )

    def aggregate_donation(self, second_party=None, exclude=None):
        """Aggregate donation data optional with exclude."""
        qs = self.queryset

        if isinstance(second_party, int):
            second_party = [second_party]

        if isinstance(exclude, int):
            exclude = [exclude]

        if exclude:
            qs = qs.exclude(first_party_id__in=exclude)

        if second_party:
            qs = qs.filter(second_party__in=second_party)

        return (
            qs.filter(Q(ref_type__in=DONATION, amount__gt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )

    def aggregate_milestone_reward(self, second_party=None):
        """Aggregate milestone reward data."""
        qs = self.queryset

        if isinstance(second_party, int):
            second_party = [second_party]

        if second_party:
            qs = qs.filter(second_party__in=second_party)

        return (
            qs.filter(Q(ref_type__in=MILESTONE_REWARD, amount__gt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )

    def aggregate_production(self, first_party=None):
        """Aggregate production data."""
        qs = self.queryset

        if isinstance(first_party, int):
            first_party = [first_party]

        if first_party:
            # Filter Costs
            qs = qs.filter(first_party__in=first_party)

        return (
            qs.filter(Q(ref_type__in=PRODUCTION) & Q(amount__gt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )

    def aggregate_traveling(self, first_party=None):
        """Aggregate traveling data."""
        qs = self.queryset

        if isinstance(first_party, int):
            first_party = [first_party]

        if first_party:
            # Filter Costs
            qs = qs.filter(first_party__in=first_party)

        return (
            qs.filter(Q(ref_type__in=TRAVELING)).aggregate(total=Sum("amount"))["total"]
            or 0
        )

    def aggregate_assets(self, first_party=None):
        """Aggregate assets cost data."""
        qs = self.queryset

        if isinstance(first_party, int):
            first_party = [first_party]

        if first_party:
            # Filter Costs
            qs = qs.filter(first_party__in=first_party)

        return (
            qs.filter(Q(ref_type__in=ASSETS) & Q(amount__lt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )

    def aggregate_skill(self, first_party=None):
        """Aggregate skill cost data."""
        qs = self.queryset

        if isinstance(first_party, int):
            first_party = [first_party]

        if first_party:
            # Filter Costs
            qs = qs.filter(first_party__in=first_party)

        return (
            qs.filter(Q(ref_type__in=SKILL) & Q(amount__lt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )

    def aggregate_planetary(self, first_party=None):
        """Aggregate planetary cost data."""
        qs = self.queryset

        if isinstance(first_party, int):
            first_party = [first_party]

        if first_party:
            # Filter Costs
            qs = qs.filter(first_party__in=first_party)

        return (
            qs.filter(Q(ref_type__in=PLANETARY) & Q(amount__lt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )

    def aggregate_rental(self, first_party=None):
        """Aggregate rental cost data."""
        qs = self.queryset

        if isinstance(first_party, int):
            first_party = [first_party]

        if first_party:
            # Filter Costs
            qs = qs.filter(first_party__in=first_party)

        return (
            qs.filter(Q(ref_type__in=RENTAL) & Q(amount__lt=0)).aggregate(
                total=Sum("amount")
            )["total"]
            or 0
        )

    def aggregate_miscellaneous(self, second_party=None):
        """Aggregate miscellaneous data."""
        qs = self.queryset

        if isinstance(second_party, int):
            second_party = [second_party]

        if second_party:
            # Filter Income
            qs = qs.filter(second_party__in=second_party)

        return (
            qs.filter(
                Q(
                    ref_type__in=[
                        *ASSETS,
                        *CONTRACT,
                        *DAILY_GOAL_REWARD,
                        *INCURSION,
                        *INSURANCE,
                        *MISSION_REWARD,
                        *MARKET,
                        *PRODUCTION,
                        *TRAVELING,
                        *SKILL,
                        *PLANETARY,
                        *LP,
                    ]
                )
                & Q(amount__gt=0)
            ).aggregate(total=Sum("amount"))["total"]
            or 0
        )



"""PvE Views"""

import logging
from datetime import datetime

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import redirect, render

# Django
from django.utils.translation import gettext_lazy as _

from allianceauth.authentication.models import CharacterOwnership

# Ledger
from ledger.api.helpers import get_corporation
from ledger.view_helpers.core import add_info_to_context

logger = logging.getLogger(__name__)


@login_required
@permission_required("ledger.advanced_access")
def corporation_ledger_index(request):
    """Corporation Ledger Index View"""
    context = {}
    context = add_info_to_context(request, context)
    return redirect(
        "ledger:corporation_ledger", request.user.profile.main_character.corporation_id
    )


@login_required
@permission_required("ledger.advanced_access")
def corporation_ledger(request, corporation_id=None):
    """
    Corporation Ledger
    """
    if corporation_id is None:
        corporation_id = request.user.profile.main_character.corporation_id

    # pylint: disable=duplicate-code
    current_year = datetime.now().year
    years = [current_year - i for i in range(6)]

    context = {
        "title": "Corporation Ledger",
        "years": years,
        "entity_pk": corporation_id,
        "corporation_id": corporation_id,
        "entity_type": "corporation",
    }
    context = add_info_to_context(request, context)
    return render(request, "ledger/corpledger/corporation_ledger.html", context=context)


@login_required
@permission_required("ledger.advanced_access")
def corporation_overview(request):
    """
    Corporation Overview
    """
    context = {"title": "Corporation Overview"}
    context = add_info_to_context(request, context)
    return render(
        request, "ledger/corpledger/admin/corporation_overview.html", context=context
    )


@login_required
@permission_required("ledger.basic_access")
def corporation_administration(request, corporation_id):
    """
    Corporation Administration
    """
    perm, corporation = get_corporation(request, corporation_id)

    if perm is False:
        msg = _("Permission Denied")
        messages.error(request, msg)
        return redirect("ledger:corporation_ledger_index")
    if perm is None:
        msg = _("Corporation not found")
        messages.info(request, msg)
        return redirect("ledger:corporation_ledger_index")
    # TODO Get Missing Characters from esi-corporations.read_corporation_membership.v1 ?
    corp_characters = CharacterOwnership.objects.filter(
        character__corporation_id=corporation.corporation.corporation_id
    ).order_by("character__character_name")

    context = {
        "corporation_id": corporation_id,
        "title": "Corporation Administration",
        "characters": corp_characters,
    }
    context = add_info_to_context(request, context)
    return render(
        request,
        "ledger/corpledger/admin/corporation_administration.html",
        context=context,
    )



"""App Tasks"""

import logging

from celery import chain, shared_task

from django.db.models import DateTimeField
from django.db.models.functions import Least
from django.utils import timezone
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.notifications import notify
from allianceauth.services.tasks import QueueOnce

from ledger import app_settings
from ledger.decorators import when_esi_is_available
from ledger.models.characteraudit import CharacterAudit
from ledger.models.corporationaudit import CorporationAudit
from ledger.models.planetary import CharacterPlanetDetails
from ledger.task_helpers.char_helpers import (
    update_character_mining,
    update_character_wallet,
)
from ledger.task_helpers.corp_helpers import update_corp_wallet_division
from ledger.task_helpers.plan_helpers import (
    update_character_planetary,
    update_character_planetary_details,
)

logger = logging.getLogger(__name__)

MAX_RETRIES_DEFAULT = 3

# Default params for all tasks.
TASK_DEFAULTS = {
    "time_limit": app_settings.LEDGER_TASKS_TIME_LIMIT,
    "max_retries": MAX_RETRIES_DEFAULT,
}

# Default params for tasks that need run once only.
TASK_DEFAULTS_ONCE = {**TASK_DEFAULTS, **{"base": QueueOnce}}

_update_ledger_planet_params = {
    **TASK_DEFAULTS_ONCE,
    **{"once": {"keys": ["character_id", "planet_id"], "graceful": True}},
}

_update_ledger_char_params = {
    **TASK_DEFAULTS_ONCE,
    **{"once": {"keys": ["character_id"], "graceful": True}},
}

_update_ledger_corp_params = {
    **TASK_DEFAULTS_ONCE,
    **{"once": {"keys": ["corporation_id"], "graceful": True}},
}


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_all_characters(runs: int = 0):
    """Update all characters"""
    characters = CharacterAudit.objects.select_related("character").all()
    for char in characters:
        update_character.apply_async(args=[char.character.character_id])
        runs = runs + 1
    logger.debug("Queued %s Character Audit Tasks", runs)


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_subset_characters(subset=5, min_runs=10, max_runs=200, force_refresh=False):
    """Update a batch of characters to prevent overload ESI"""
    total_characters = CharacterAudit.objects.filter(active=1).count()
    characters_count = min(max(total_characters // subset, min_runs), total_characters)

    # Limit the number of characters to update to prevent overload ESI
    characters_count = min(characters_count, max_runs)

    # Annotate characters with the oldest update timestamp and order by it
    characters = (
        CharacterAudit.objects.filter(active=1)
        .annotate(
            oldest_update=Least(
                "last_update_wallet",
                "last_update_mining",
                "last_update_planetary",
                output_field=DateTimeField(),
            )
        )
        .order_by("oldest_update")[:characters_count]
    )

    for char in characters:
        update_character.apply_async(
            args=[char.character.character_id], force_refresh=force_refresh
        )
    logger.debug("Queued %s Character Audit Tasks", len(characters))


@shared_task(**_update_ledger_char_params)
@when_esi_is_available
def update_character(character_id: int, force_refresh=False):
    character = CharacterAudit.objects.get(character__character_id=character_id)

    # Settings for Task Queue
    skip_date = timezone.now() - timezone.timedelta(
        minutes=app_settings.LEDGER_STALE_STATUS
    )
    que = []
    mindt = timezone.now() - timezone.timedelta(days=7)
    priority = 7

    logger.debug(
        "Processing Audit Updates for %s", format(character.character.character_name)
    )

    if (character.last_update_mining or mindt) <= skip_date or force_refresh:
        que.append(
            update_char_mining_ledger.si(
                character.character.character_id, force_refresh=force_refresh
            ).set(priority=priority)
        )

    if (character.last_update_wallet or mindt) <= skip_date or force_refresh:
        que.append(
            update_char_wallet.si(
                character.character.character_id, force_refresh=force_refresh
            ).set(priority=priority)
        )

    if (character.last_update_planetary or mindt) <= skip_date or force_refresh:
        que.append(
            update_char_planets.si(
                character.character.character_id, force_refresh=force_refresh
            ).set(priority=priority)
        )

    chain(que).apply_async()
    logger.info(
        "Queued %s Audit Updates for %s",
        len(que),
        character.character.character_name,
    )


@shared_task(**_update_ledger_char_params)
def update_char_mining_ledger(character_id, force_refresh=False):
    return update_character_mining(character_id, force_refresh=force_refresh)


@shared_task(**_update_ledger_char_params)
def update_char_wallet(
    character_id, force_refresh=False
):  # pylint: disable=unused-argument, dangerous-default-value
    return update_character_wallet(character_id, force_refresh=force_refresh)


@shared_task(**_update_ledger_char_params)
def update_char_planets(
    character_id, force_refresh=False
):  # pylint: disable=unused-argument, dangerous-default-value
    return update_character_planetary(character_id, force_refresh=force_refresh)


@shared_task(**_update_ledger_planet_params)
def update_char_planets_details(
    character_id, planet_id, force_refresh=False
):  # pylint: disable=unused-argument, dangerous-default-value
    return update_character_planetary_details(
        character_id, planet_id, force_refresh=force_refresh
    )


# pylint: disable=unused-argument, too-many-locals
@shared_task(**TASK_DEFAULTS_ONCE)
def check_planetary_alarms(runs: int = 0):
    all_planets = CharacterPlanetDetails.objects.all()
    owner_ids = {}
    warnings = {}

    for planet in all_planets:
        if planet.is_expired and not planet.notification_sent and planet.notification:
            character_id = planet.planet.character.character.character_id

            # Determine if the character_id is part of any main character's alts
            main_id = None
            for main, alts in owner_ids.items():
                if character_id in alts:
                    main_id = main
                    break

            if main_id is None:
                try:
                    owner = CharacterOwnership.objects.get(
                        character__character_id=character_id
                    )
                    main = owner.user.profile.main_character
                    alts = main.character_ownership.user.character_ownerships.all()

                    owner_ids[main.character_id] = alts.values_list(
                        "character__character_id", flat=True
                    )

                    main_id = main.character_id
                except CharacterOwnership.DoesNotExist:
                    continue
                except AttributeError:
                    continue

            msg = _("%(charname)s on %(planetname)s") % {
                "charname": planet.planet.character.character.character_name,
                "planetname": planet.planet.planet.name,
            }

            if main_id not in warnings:
                warnings[main_id] = []

            warnings[main_id].append(msg)
            planet.notification_sent = True
            planet.save()

    if warnings:
        for main_id, messages in warnings.items():
            owner = CharacterOwnership.objects.get(character__character_id=main_id)
            title = _("Planetary Extractor Heads Expired")

            # Split messages into chunks of 50
            for i in range(0, len(messages), 50):
                chunk = messages[i : i + 50]
                msg = "\n".join(chunk)

                full_message = format_html(
                    "Following Planet Extractor Heads have expired: \n{}", msg
                )
                notify(
                    title=title,
                    message=full_message,
                    user=owner.user,
                    level="warning",
                )
                runs = runs + 1
    logger.info("Queued %s Planetary Alarms.", runs)


# Corporation Audit - Tasks
@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_all_corps(runs: int = 0):
    corps = CorporationAudit.objects.select_related("corporation").all()
    for corp in corps:
        update_corp.apply_async(args=[corp.corporation.corporation_id])
        runs = runs + 1
    logger.debug("Queued %s Corporation Audit Tasks", runs)


@shared_task(**_update_ledger_corp_params)
@when_esi_is_available
def update_corp(corporation_id, force_refresh=False):  # pylint: disable=unused-argument
    corp = CorporationAudit.objects.get(corporation__corporation_id=corporation_id)
    logger.debug("Processing Audit Updates for %s", corp.corporation.corporation_name)

    # Settings for Task Queue
    skip_date = timezone.now() - timezone.timedelta(
        minutes=app_settings.LEDGER_STALE_STATUS
    )
    mindt = timezone.now() - timezone.timedelta(days=7)
    priority = 7

    if (corp.last_update_wallet or mindt) <= skip_date or force_refresh:
        update_corp_wallet.si(corporation_id, force_refresh=force_refresh).set(
            priority=priority
        ).apply_async()

    logger.info("Queued Audit Updates for %s", corp.corporation.corporation_name)


@shared_task(**_update_ledger_corp_params)
def update_corp_wallet(corporation_id, force_refresh=False):
    return update_corp_wallet_division(corporation_id, force_refresh=force_refresh)



from django.contrib.auth.models import User

from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import (
    create_user_from_evecharacter,
)

from ledger.models.corporationaudit import CorporationAudit
from ledger.tests.testdata.generate_characteraudit import (
    add_auth_character_to_user,
    create_user_from_evecharacter_with_access,
)


def create_corporationaudit_from_evecharacter(
    eve_character: EveCharacter, **kwargs
) -> CorporationAudit:
    """Create a LedgerAudit Corporation from EveCharacter"""
    params = {
        "corporation_name": eve_character.corporation_name,
        "corporation": eve_character.corporation,
    }
    params.update(kwargs)
    corporation = CorporationAudit(**params)
    corporation.save()
    return corporation


def create_corporationaudit_character(character_id: int, **kwargs) -> CorporationAudit:
    """Create a Audit Character from a existing EveCharacter"""

    _, character_ownership = create_user_from_evecharacter_with_access(
        character_id, disconnect_signals=True
    )
    return create_corporationaudit_from_evecharacter(
        character_ownership.character, **kwargs
    )


def add_corporationaudit_corporation_to_user(
    user: User, character_id: int, disconnect_signals: bool = True, **kwargs
) -> CorporationAudit:
    character_ownership = add_auth_character_to_user(
        user,
        character_id,
        disconnect_signals=disconnect_signals,
    )
    return create_corporationaudit_from_evecharacter(
        character_ownership.character, **kwargs
    )



import logging

from django.db import models

logger = logging.getLogger(__name__)


class CorpAuditQuerySet(models.QuerySet):
    def visible_to(self, user):
        # superusers get all visible
        if user.is_superuser:
            logger.debug(
                "Returning all corps for superuser %s.",
                user,
            )
            return self

        if user.has_perm("ledger.corp_audit_admin_manager"):
            logger.debug("Returning all corps for Corp Audit Admin %s.", user)
            return self

        try:
            char = user.profile.main_character
            assert char
            query = None

            if user.has_perm("ledger.advanced_access"):
                query = models.Q(corporation__corporation_id=char.corporation_id)

            logger.debug("Returning own corps for User %s.", user)

            if query is None:
                return self.none()

            return self.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return self.none()


class CorpAuditManagerBase(models.Manager):
    def visible_to(self, user):
        return self.get_queryset().visible_to(user)


CorpAuditManager = CorpAuditManagerBase.from_queryset(CorpAuditQuerySet)



import logging

from ninja import NinjaAPI

from django.core.handlers.wsgi import WSGIRequest
from django.shortcuts import render
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from allianceauth.eveonline.models import EveCharacter

from ledger.api import schema
from ledger.api.api_helper.alliance_helper import AllianceProcess
from ledger.api.api_helper.character_helper import CharacterProcess
from ledger.api.api_helper.corporation_helper import (
    CorporationProcess,
)
from ledger.api.helpers import (
    get_alliance,
    get_alts_queryset,
    get_character,
    get_corporation,
)

logger = logging.getLogger(__name__)


def ledger_api_process(
    request, entity_type: str, entity_id: int, date: timezone.datetime, view: str
):
    singleview = request.GET.get("single", False)
    perm = True
    result = {"perm": None, "process": None}

    if entity_type == "character":
        character_id = request.GET.get("character_id", None)
        perm, character = get_character(request, entity_id)

        if character_id is None and singleview is False:
            characters = get_alts_queryset(character)
        elif character_id is None:
            characters = EveCharacter.objects.filter(
                character_id=character.character_id,
            )
        else:
            perm, character = get_character(request, character_id)
            characters = EveCharacter.objects.filter(
                character_id=character_id,
            )

        if character is not None:
            result["perm"] = perm
            result["process"] = CharacterProcess(
                main=character, chars=characters, date=date, view=view
            )

    elif entity_type == "corporation":
        main_character_id = request.GET.get("main_character_id", None)
        perm, corporation = get_corporation(request, entity_id)
        main_character = None

        if corporation is not None:
            if main_character_id:
                main_character = get_character(request, main_character_id)[1]

            result["perm"] = perm
            result["process"] = CorporationProcess(
                corporation=corporation,
                date=date,
                main_character=main_character,
                view=view,
            )

    elif entity_type == "alliance":
        corporation_id = request.GET.get("corporation_id", None)
        perm, alliance = get_alliance(request, entity_id)
        corporation = None

        if alliance is not None:
            if corporation_id:
                corporation = get_corporation(request, corporation_id)[1]

            result["perm"] = perm
            result["process"] = AllianceProcess(
                alliance=alliance, corporation=corporation, date=date, view=view
            )

    return result["perm"], result["process"]


class LedgerApiEndpoints:
    tags = ["Ledger"]

    def __init__(self, api: NinjaAPI):
        @api.get(
            "{entity_type}/{entity_id}/ledger/date/{date}/view/{view}/",
            response={200: schema.Ledger, 403: str, 404: str},
            tags=self.tags,
        )
        def get_ledger(
            request: WSGIRequest, entity_type: str, entity_id: int, date: str, view: str
        ):
            try:
                date_obj = timezone.datetime.strptime(date, "%Y-%m-%d").date()
            except ValueError:
                return 403, "Invalid Date format. Use YYYY-MM-DD"

            perm, ledger = ledger_api_process(
                request, entity_type, entity_id, date_obj, view
            )

            if perm is False:
                return 403, str(_("Permission Denied"))
            if perm is None:
                return 404, str(_("Entity Not Found"))

            output = ledger.generate_ledger()
            return output

        @api.get(
            "{entity_type}/{entity_id}/template/date/{date}/view/{view}/",
            response={200: schema.Ledger, 403: str, 404: str},
            tags=self.tags,
        )
        def get_entity_information(
            request: WSGIRequest, entity_type: str, entity_id: int, date: str, view: str
        ):
            try:
                date_obj = timezone.datetime.strptime(date, "%Y-%m-%d").date()
            except ValueError:
                return 403, "Invalid Date format. Use YYYY-MM-DD"

            perm, ledger = ledger_api_process(
                request, entity_type, entity_id, date_obj, view
            )

            if perm is False:
                return 403, str(_("Permission Denied"))
            if perm is None:
                return 404, str(_("Entity Not Found"))

            context = {
                "character": ledger.generate_template(),
                "mode": "CHARACTER",
            }
            return render(
                request,
                "ledger/partials/information/view_character_content.html",
                context,
            )



"""
Planetary Helpers
"""

import logging

from django.utils import timezone
from eveuniverse.models import EvePlanet

from ledger.decorators import log_timing
from ledger.models.characteraudit import CharacterAudit
from ledger.models.planetary import CharacterPlanet, CharacterPlanetDetails
from ledger.providers import esi
from ledger.task_helpers.core_helpers import get_token
from ledger.task_helpers.etag_helpers import (
    HTTPGatewayTimeoutError,
    NotModifiedError,
    etag_results,
)

logger = logging.getLogger(__name__)


def convert_datetime_to_str(data):
    if isinstance(data, dict):
        return {k: convert_datetime_to_str(v) for k, v in data.items()}
    if isinstance(data, list):
        return [convert_datetime_to_str(i) for i in data]
    if isinstance(data, timezone.datetime):
        return data.isoformat()
    return data


# pylint: disable=too-many-locals
@log_timing(logger)
def update_character_planetary(character_id, force_refresh=False):
    # pylint: disable=import-outside-toplevel, cyclic-import
    from ledger.tasks import update_char_planets_details

    audit_char = CharacterAudit.objects.get(character__character_id=character_id)
    logger.debug("Updating planets for: %s", audit_char.character.character_name)

    token = get_token(character_id, CharacterPlanet.get_esi_scopes())

    if not token:
        logger.info(
            "No Planet Token for %s, Deactivate Character",
            audit_char.character.character_name,
        )
        audit_char.is_active()
        return "No Tokens"

    try:
        planet_items_data = (
            esi.client.Planetary_Interaction.get_characters_character_id_planets(
                character_id=character_id
            )
        )

        planet_items = etag_results(
            planet_items_data, token, force_refresh=force_refresh
        )

        _current_planets = CharacterPlanet.objects.filter(
            character=audit_char
        ).values_list("planet_id", flat=True)

        _planets_ids = []
        _planets_new = []
        _planets_update = []

        for planet in planet_items:
            eve_planet, _ = EvePlanet.objects.get_or_create_esi(id=planet["planet_id"])
            _planets_ids.append(eve_planet.id)

            try:
                e_planet = CharacterPlanet.objects.get(
                    character=audit_char, planet=eve_planet
                )
                prim_key = e_planet.id
            except CharacterPlanet.DoesNotExist:
                prim_key = None

            planet = CharacterPlanet(
                id=prim_key,
                character=audit_char,
                planet=eve_planet,
                planet_name=eve_planet.name,
                upgrade_level=planet["upgrade_level"],
                num_pins=planet["num_pins"],
                last_update=timezone.now(),
            )

            if eve_planet.id not in _current_planets:
                _planets_new.append(planet)
            else:
                _planets_update.append(planet)

        if _planets_new:
            CharacterPlanet.objects.bulk_create(_planets_new)
        if _planets_update:
            CharacterPlanet.objects.bulk_update(
                _planets_update, fields=["upgrade_level", "num_pins", "last_update"]
            )

        # Delete Planets that are no longer in the list
        obsolete_planets = set(_current_planets) - set(_planets_ids)
        CharacterPlanet.objects.filter(
            character=audit_char, planet_id__in=obsolete_planets
        ).delete()
        # Delete Planet Details that are no longer in the list
        CharacterPlanetDetails.objects.filter(
            planet__character=audit_char, planet__planet_id__in=obsolete_planets
        ).delete()

        for planet in CharacterPlanet.objects.filter(character=audit_char):
            update_char_planets_details.apply_async(
                args=[character_id, planet.planet_id],
                kwargs={"force_refresh": False},
                priority=8,
            )

        logger.debug(
            "Finished planets update for: %s", audit_char.character.character_name
        )
    except NotModifiedError:
        logger.debug("No New Planet data for: %s", audit_char.character.character_name)
    except HTTPGatewayTimeoutError:
        logger.debug("Gateway Timeout for: %s", audit_char.character.character_name)
        return "Gateway Timeout"

    audit_char.last_update_planetary = timezone.now()
    audit_char.save()
    audit_char.is_active()
    return ("Finished planets update for: %s", audit_char.character.character_name)


@log_timing(logger)
def update_character_planetary_details(character_id, planet_id, force_refresh=False):
    planet_char = CharacterPlanet.objects.get(
        character__character__character_id=character_id, planet__id=planet_id
    )

    logger.debug(
        "Updating planet details %s for: %s",
        planet_char.planet.name,
        planet_char.character.character.character_name,
    )

    token = get_token(character_id, CharacterPlanet.get_esi_scopes())

    if not token:
        return "No Tokens"

    try:
        planet_details_data = esi.client.Planetary_Interaction.get_characters_character_id_planets_planet_id(
            character_id=character_id, planet_id=planet_id
        )

        planet_details_items = etag_results(
            planet_details_data, token, force_refresh=force_refresh
        )
        # Convert Time to string
        planet_details_items = convert_datetime_to_str(planet_details_items)

        planet, created = CharacterPlanetDetails.objects.update_or_create_layout(
            planet=planet_char,
            esi_result=planet_details_items,
        )

        if not created:
            # Set Alert if Extractor Heads are expired
            if planet.is_expired and planet.last_alert is None:
                logger.debug(
                    "Planet %s Extractor Heads Expired for: %s",
                    planet.planet.planet.name,
                    planet.planet.character.character.character_name,
                )
                planet.last_alert = timezone.now()

            # Reset Alert after 1 day
            if (
                planet.last_alert is not None
                and planet.last_alert < timezone.now() - timezone.timedelta(days=1)
            ):
                logger.debug(
                    "Notification Reseted for %s Planet: %s",
                    planet.planet.character.character.character_name,
                    planet.planet.planet.name,
                )
                planet.last_alert = None
                planet.notification_sent = False

            planet.save()
        logger.debug(
            "Finished planets details update for: %s",
            planet_char.character.character.character_name,
        )
        return (
            "Finished planets details update for: %s",
            planet_char.character.character.character_name,
        )
    except NotModifiedError:
        logger.debug(
            "No New Planet Details data for: %s",
            planet_char.character.character.character_name,
        )
        return (
            "No New Planet Details data for: %s",
            planet_char.character.character.character_name,
        )
    except HTTPGatewayTimeoutError:
        logger.debug(
            "Gateway Timeout for: %s", planet_char.character.character.character_name
        )
        return (
            "Gateway Timeout for: %s",
            planet_char.character.character.character_name,
        )



"""PvE Views"""

import logging

# Django
from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import redirect

# Ledger
from ledger.view_helpers.core import add_info_to_context

logger = logging.getLogger(__name__)


@login_required
@permission_required("ledger.basic_access")
def index(request):
    """Index View"""
    context = {
        "title": "Ledger",
    }
    context = add_info_to_context(request, context)
    return redirect(
        "ledger:character_ledger", request.user.profile.main_character.character_id
    )



import logging
from collections import defaultdict
from decimal import Decimal

from django.db.models import Q
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from allianceauth.eveonline.models import EveAllianceInfo

from ledger.api.api_helper.aggregator import AggregateLedger
from ledger.api.api_helper.billboard_helper import BillboardSystem
from ledger.api.api_helper.information_helper import (
    InformationData,
)
from ledger.models.corporationaudit import (
    CorporationAudit,
    CorporationWalletJournalEntry,
)
from ledger.view_helpers.core import events_filter

logger = logging.getLogger(__name__)


# pylint: disable=duplicate-code
class AllianceProcess:
    """JournalProcess class to process the journal entries."""

    def __init__(
        self,
        alliance: EveAllianceInfo,
        date: timezone.datetime,
        corporation: CorporationAudit = None,
        view=None,
    ):
        self.alliance = alliance
        self.corporation = corporation
        self.date = date
        self.view = view
        self.current_date = timezone.now()

        self._init_journal()

    def _filter_date(self):
        """Filter the date."""
        filter_date = Q(date__year=self.date.year)
        if self.view == "month":
            filter_date &= Q(date__month=self.date.month)
        elif self.view == "day":
            filter_date &= Q(date__month=self.date.month)
            filter_date &= Q(date__day=self.date.day)
        return filter_date

    def _init_journal(self):
        """Get the involved characters for the alliance"""
        self.corporation_journal = CorporationWalletJournalEntry.objects.filter(
            self._filter_date(),
            division__corporation__corporation__alliance__alliance_id=self.alliance.alliance_id,
        )

        # Exclude Corp Tax Events
        self.corporation_journal = events_filter(self.corporation_journal)

        # Get Glances
        self.glance = AggregateLedger(self.corporation_journal)

        # Get Entity IDs from Corporation or Alliance
        if self.corporation:
            journal = self.corporation_journal.filter(
                Q(division__corporation=self.corporation)
            )
            self.entity_ids = set(
                journal.values_list("second_party_id", flat=True)
            ) | set(journal.values_list("first_party_id", flat=True))
        else:
            self.entity_ids = set(
                self.corporation_journal.values_list("second_party_id", flat=True)
            ) | set(self.corporation_journal.values_list("first_party_id", flat=True))

        self.glance_day = self.corporation_journal.filter(
            date__year=self.current_date.year,
            date__month=self.current_date.month,
            date__day=self.current_date.day,
        )

    def _process_journal_entry(self, entry, main_dict, billboard):
        """Process a single journal entry."""
        corp_id = entry["division__corporation__corporation__corporation_id"]
        corp_name = entry["division__corporation__corporation__corporation_name"]

        bounty = entry.get("bounty_income", 0)
        ess = entry.get("ess_income", 0)
        miscellaneous = entry.get("miscellaneous", 0)
        costs = entry.get("costs", 0) + self.glance.aggregate_corp_withdraw_cost(
            exclude=corp_id
        )

        if bounty > 0 or ess > 0 or costs > 0 or miscellaneous > 0:
            main_dict[corp_id] = {
                "main_id": corp_id,
                "main_name": corp_name,
                "entity_type": "corporation",
                "total_amount": bounty,
                "total_amount_ess": ess,
                "total_amount_costs": costs,
                "total_amount_others": miscellaneous,
            }

            billboard.chord_add_data(
                corp_name, _("Income"), (bounty + ess + miscellaneous)
            )

    def generate_ledger(self):
        """Generate the ledger for the alliance"""
        main_dict = {}
        billboard = BillboardSystem(view=self.view)
        exclude = self.corporation_journal.values_list(
            "division__corporation__corporation__corporation_id", flat=True
        )

        journal = (
            self.corporation_journal.values(
                "division__corporation__corporation__corporation_id",
                "division__corporation__corporation__corporation_name",
            )
            .annotate_bounty_income()
            .annotate_ess_income()
            .annotate_miscellaneous()
            .annotate_costs()
        )

        for entry in journal:
            self._process_journal_entry(entry, main_dict, billboard)

        totals = {
            "total_amount": self.glance.aggregate_bounty(),
            "total_amount_ess": self.glance.aggregate_ess(),
            "total_amount_others": (
                +self.glance.aggregate_miscellaneous()
                + self.glance.aggregate_donation()
                + self.glance.aggregate_corp_withdraw(
                    exclude=exclude
                )  # Exclude Intern Transfers
            ),
            "total_amount_costs": self.glance.aggregate_costs()
            + self.glance.aggregate_corp_withdraw_cost(exclude=exclude),
        }
        totals["total_amount_all"] = (
            totals["total_amount"]
            + totals["total_amount_ess"]
            + totals["total_amount_others"]
            + totals["total_amount_costs"]
        )

        # Add Rattingbar for the corporation
        rattingbar_timeline = billboard.create_timeline(self.corporation_journal)
        rattingbar = (
            rattingbar_timeline.annotate_bounty_income()
            .annotate_ess_income()
            .annotate_miscellaneous()
        )
        billboard.create_or_update_results(rattingbar)
        billboard.create_ratting_bar()

        # Order and Handle Overflow
        billboard.chord_handle_overflow()

        output = {
            "ratting": sorted(list(main_dict.values()), key=lambda x: x["main_name"]),
            "billboard": billboard.dict,
            "total": totals,
        }

        return output

    def generate_template(self):
        """Generate the information for the corporation"""
        information_dict = {}
        exclude = None

        # Create the Ledger
        ledger_data = InformationData(
            corporation=self.corporation,
            date=self.date,
            view=self.view,
            current_date=self.current_date,
        )

        if self.corporation is not None:
            exclude = self.corporation.corporation.corporation_id
        else:
            exclude = self.corporation_journal.values_list(
                "division__corporation__corporation__corporation_id", flat=True
            )

        day_aggregate = AggregateLedger(self.glance_day)

        amounts = defaultdict(lambda: defaultdict(Decimal))

        amounts["bounty_income"] = {
            "total_amount": self.glance.aggregate_bounty(self.entity_ids),
            "total_amount_day": day_aggregate.aggregate_bounty(self.entity_ids),
        }

        amounts["ess_income"] = {
            "total_amount": self.glance.aggregate_ess(self.entity_ids),
            "total_amount_day": day_aggregate.aggregate_ess(self.entity_ids),
        }

        amounts["mission_income"] = {
            "total_amount": self.glance.aggregate_mission(self.entity_ids),
            "total_amount_day": day_aggregate.aggregate_mission(self.entity_ids),
        }

        amounts["market_income"] = {
            "total_amount": self.glance.aggregate_market(second_party=self.entity_ids),
            "total_amount_day": day_aggregate.aggregate_market(
                second_party=self.entity_ids
            ),
        }

        amounts["contract_income"] = {
            "total_amount": self.glance.aggregate_contract(self.entity_ids),
            "total_amount_day": day_aggregate.aggregate_contract(self.entity_ids),
        }

        amounts["donation_income"] = {
            "total_amount": self.glance.aggregate_donation(self.entity_ids)
            + self.glance.aggregate_corp_withdraw(self.entity_ids, exclude=exclude),
            "total_amount_day": day_aggregate.aggregate_donation(self.entity_ids)
            + day_aggregate.aggregate_corp_withdraw(self.entity_ids, exclude=exclude),
        }

        amounts["insurance_income"] = {
            "total_amount": self.glance.aggregate_insurance(self.entity_ids),
            "total_amount_day": day_aggregate.aggregate_insurance(self.entity_ids),
        }

        amounts["daily_goal_income"] = {
            "total_amount": self.glance.aggregate_daily_goal(self.entity_ids),
            "total_amount_day": day_aggregate.aggregate_daily_goal(self.entity_ids),
        }

        amounts["citadel_income"] = {
            "total_amount": (
                self.glance.aggregate_production(self.entity_ids)
                + self.glance.aggregate_traveling(self.entity_ids)
            ),
            "total_amount_day": (
                day_aggregate.aggregate_production(self.entity_ids)
                + day_aggregate.aggregate_traveling(self.entity_ids)
            ),
        }

        amounts["market_cost"] = {
            "total_amount": self.glance.aggregate_market_cost(self.entity_ids),
            "total_amount_day": day_aggregate.aggregate_market_cost(self.entity_ids),
        }

        amounts["rental_cost"] = {
            "total_amount": self.glance.aggregate_rental(self.entity_ids),
            "total_amount_day": day_aggregate.aggregate_rental(self.entity_ids),
        }

        amounts["donation_cost"] = {
            "total_amount": self.glance.aggregate_corp_withdraw_cost(
                self.entity_ids, exclude=exclude
            ),
            "total_amount_day": day_aggregate.aggregate_corp_withdraw_cost(
                self.entity_ids, exclude=exclude
            ),
        }

        amounts["contract_cost"] = {
            "total_amount": self.glance.aggregate_contract_cost(self.entity_ids),
            "total_amount_day": day_aggregate.aggregate_contract_cost(self.entity_ids),
        }

        amounts["production_cost"] = {
            "total_amount": self.glance.aggregate_production_cost(self.entity_ids),
            "total_amount_day": day_aggregate.aggregate_production_cost(
                self.entity_ids
            ),
        }

        information_dict.update(
            {
                "main_id": ledger_data.id,
                "main_name": ledger_data.name,
                "date": ledger_data.information_date,
            }
        )
        information_dict = ledger_data._generate_amounts_dict(amounts, information_dict)
        return information_dict



import logging
from dataclasses import dataclass
from typing import Any

from django.db.models import QuerySet
from django.db.models.functions import TruncDay, TruncHour, TruncMonth

logger = logging.getLogger(__name__)


@dataclass
class ChartData:
    title: str
    categories: list[str]
    series: list[dict[str, Any]]


class BillboardSystem:
    """BillboardSystem class to process billboard data."""

    @dataclass
    class BillboardDict:
        """BillboardDict class to store the billboard data."""

        charts: ChartData = None
        rattingbar: ChartData = None
        workflowgauge: ChartData = None

    def __init__(
        self,
        view,
    ):
        self.view = view
        self.dict = self.BillboardDict()
        self.results = {}

    def _get_formatted_date(self, date, view):
        if view == "year":
            return date.strftime("%Y-%m")
        if view == "month":
            return date.strftime("%Y-%m-%d")
        if self.view == "day":
            return date.strftime("%Y-%m-%d %H:%M")
        raise ValueError("Invalid view type. Use 'day', 'month', or 'year'.")

    def _create_chart_dict(self):
        """Create the Charts dict if it doesn't exist"""
        if self.dict.charts is None:
            self.dict.charts = ChartData(
                title="Billboard",
                categories=[],
                series=[],
            )

    def chord_add_char_data_from_dict(self, data: dict):
        """Add character data to chord from dict"""
        self._create_chart_dict()

        data_points = [
            {
                "from": f"{data['main_name']}",
                "to": "Wallet",
                "value": abs(data["total_amount"]),
            },
            {
                "from": f"{data['main_name']}",
                "to": "Wallet",
                "value": abs(data["total_amount_ess"]),
            },
            {
                "from": f"{data['main_name']}",
                "to": "Wallet",
                "value": abs(data["total_amount_mining"]),
            },
            {
                "from": f"{data['main_name']}",
                "to": "Wallet",
                "value": abs(data["total_amount_others"]),
            },
            {
                "from": f"{data['main_name']}",
                "to": "Costs",
                "value": abs(data["total_amount_costs"]),
            },
        ]

        for point in data_points:
            if point["value"] != 0:
                self.dict.charts.series.append(point)

    def chord_add_data(self, chord_from: str, chord_to: str, value: int):
        """Add Simple Chord data"""
        self._create_chart_dict()

        if value == 0:
            return

        data = {
            "from": chord_from,
            "to": chord_to,
            "value": value,
        }
        self.dict.charts.series.append(data)

    def chord_handle_overflow(self):
        """Order and handle overflow data for the billboard"""
        if self.dict.charts is None:
            return

        self.dict.charts.series = sorted(
            self.dict.charts.series, key=lambda x: x["value"], reverse=True
        )
        if len(self.dict.charts.series) > 10:
            others_value = sum(entry["value"] for entry in self.dict.charts.series[10:])
            self.dict.charts.series = self.dict.charts.series[:10]
            self.dict.charts.series.append(
                {
                    "from": "Others",
                    "to": "Wallet",
                    "value": others_value,
                }
            )

    # TODO Add Mining to the billboard
    def create_timeline(self, journal: QuerySet):
        """Create the timeline data for the billboard"""
        qs = journal

        if self.view == "year":
            qs = qs.annotate(period=TruncMonth("date"))
        elif self.view == "month":
            qs = qs.annotate(period=TruncDay("date"))
        elif self.view == "day":
            qs = qs.annotate(period=TruncHour("date"))
        else:
            raise ValueError("Invalid view type. Use 'day', 'month', or 'year'.")

        qs = qs.values("period").order_by("period")
        return qs

    def create_or_update_results(self, qs: QuerySet[dict]):
        """Create or update the results for the billboard"""
        for entry in qs:
            date = entry["period"]
            bounty = entry.get("bounty_income", 0)
            ess = entry.get("ess_income", 0)
            miscellaneous = entry.get("miscellaneous", 0)

            # Store the results in a dictionary
            if date not in self.results:
                self.results[date] = {
                    "bounty": 0,
                    "ess": 0,
                    "miscellaneous": 0,
                }
            self.results[date]["bounty"] += bounty
            self.results[date]["ess"] += ess
            self.results[date]["miscellaneous"] += miscellaneous

        return self.results

    def create_ratting_bar(self):
        """Create the ratting bar data for the billboard"""
        formatted_results = []

        for date, values in self.results.items():
            # Remove categories with value 0
            filtered_values = {k: v for k, v in values.items() if v != 0}
            if not filtered_values:
                continue  # Skip if all categories are 0

            formatted_results.append(
                {
                    "date": self._get_formatted_date(date, self.view),
                    **{k: int(v) for k, v in filtered_values.items()},
                }
            )

        if not formatted_results:
            return []

        self.dict.rattingbar = ChartData(
            title="Ratting Bar",
            categories=["Bounty", "ESS", "Miscellaneous"],
            series=formatted_results,
        )
        return formatted_results



"""
Core Helpers
"""

import logging

from esi.models import Token

logger = logging.getLogger(__name__)


def get_token(character_id: int, scopes: list) -> Token:
    """
    Helper method to get a valid token for a specific character with specific scopes.

    Parameters
    ----------
    character_id: `int`
    scopes: `int`

    Returns
    ----------
    `class`: esi.models.Token or False

    """
    token = (
        Token.objects.filter(character_id=character_id)
        .require_scopes(scopes)
        .require_valid()
        .first()
    )
    if token:
        return token
    return False



import logging
from collections import defaultdict
from decimal import Decimal

from django.db.models import Q, QuerySet
from django.utils import timezone

from allianceauth.eveonline.models import EveCharacter

from ledger.api.api_helper.aggregator import AggregateLedger, AggregateMining
from ledger.api.api_helper.billboard_helper import BillboardSystem
from ledger.api.api_helper.information_helper import (
    InformationData,
)
from ledger.api.helpers import get_alts_queryset
from ledger.models.characteraudit import (
    CharacterMiningLedger,
    CharacterWalletJournalEntry,
)
from ledger.models.corporationaudit import (
    CorporationWalletJournalEntry,
)
from ledger.view_helpers.core import events_filter

logger = logging.getLogger(__name__)


class CharacterGlances:
    """CharacterGlances class to process the glances for the character."""

    def __init__(
        self, char_journal: QuerySet, corp_journal: QuerySet, mining_journal: QuerySet
    ):
        self.glance_char = AggregateLedger(char_journal)
        self.glance_corp = AggregateLedger(corp_journal)
        self.glance_mining = AggregateMining(mining_journal)
        self.glance_day_char = char_journal.filter(
            date__year=timezone.now().year,
            date__month=timezone.now().month,
            date__day=timezone.now().day,
        )
        self.glance_day_corp = corp_journal.filter(
            date__year=timezone.now().year,
            date__month=timezone.now().month,
            date__day=timezone.now().day,
        )
        self.glance_day_mining = mining_journal.filter(
            date__year=timezone.now().year,
            date__month=timezone.now().month,
            date__day=timezone.now().day,
        )


# pylint: disable=too-many-instance-attributes
class CharacterProcess:
    """JournalProcess class to process the journal entries."""

    def __init__(
        self,
        main: EveCharacter,
        chars: QuerySet[EveCharacter],
        date: timezone.datetime,
        view=None,
    ):
        self.main = main
        self.chars = chars
        self.date = date
        self.view = view
        self.current_date = timezone.now()

        self._init_journal()

    def _filter_date(self):
        # pylint: disable=duplicate-code
        filter_date = Q(date__year=self.date.year)
        if self.view == "month":
            filter_date &= Q(date__month=self.date.month)
        elif self.view == "day":
            filter_date &= Q(date__month=self.date.month)
            filter_date &= Q(date__day=self.date.day)
        return filter_date

    def _init_journal(self):
        """Initialize the data for the ledger"""
        self.chars_ids = [char.character_id for char in self.chars]
        self.alt_ids = self.chars_ids
        if len(self.chars) == 1:
            self.alt_ids = get_alts_queryset(self.main).values_list(
                "character_id", flat=True
            )

        self.character_journal = CharacterWalletJournalEntry.objects.filter(
            self._filter_date(), character__character__character_id__in=self.chars_ids
        )
        self.mining_journal = CharacterMiningLedger.objects.filter(
            self._filter_date(), character__character__character_id__in=self.chars_ids
        )
        self.corporation_journal = self._init_corporation_journal(self.chars_ids)

        self.glances = CharacterGlances(
            char_journal=self.character_journal,
            corp_journal=self.corporation_journal,
            mining_journal=self.mining_journal,
        )

    def _init_corporation_journal(self, chars_ids: list[int]):
        """Get the involved characters for the corporation"""
        corporation_journal = CorporationWalletJournalEntry.objects.filter(
            self._filter_date(),
            Q(first_party_id__in=chars_ids) | Q(second_party_id__in=chars_ids),
        )

        # Exclude Corp Tax Events
        corporation_journal = events_filter(corporation_journal)

        return corporation_journal

    def generate_ledger(self):
        ratting = {}
        billboard = BillboardSystem(self.view)

        def process_character(char):
            """Helper method to process a single character."""
            bounty = self.glances.glance_char.aggregate_bounty(char.character_id)
            ess = self.glances.glance_corp.aggregate_ess(
                char.character_id, is_character=True
            )
            mining = self.glances.glance_mining.aggregate_mining(char.character_id)
            miscellaneous = self.glances.glance_char.aggregate_miscellaneous(
                char.character_id
            )
            donation = self.glances.glance_char.aggregate_donation(
                char.character_id, exclude=self.alt_ids
            )
            daily_goal = self.glances.glance_corp.aggregate_daily_goal(
                char.character_id, is_character=True
            )
            miscellaneous += donation + daily_goal
            costs = self.glances.glance_char.aggregate_costs(char.character_id)

            if bounty == 0 and ess == 0 and miscellaneous == 0 and costs == 0:
                return None

            data = {
                "main_id": char.character_id,
                "main_name": char.character_name,
                "entity_type": "character",
                "total_amount": bounty,
                "total_amount_ess": ess,
                "total_amount_mining": mining,
                "total_amount_others": miscellaneous,
                "total_amount_costs": costs,
            }
            return data

        for char in self.chars:
            data = process_character(char)
            if data:
                ratting[char.character_id] = data
                billboard.chord_add_char_data_from_dict(data)

        # Add Character Journal to the billboard
        rattingbar_timeline = billboard.create_timeline(self.character_journal)
        rattingbar = rattingbar_timeline.annotate_bounty_income().annotate_miscellaneous_with_exclude(
            exclude=self.alt_ids
        )
        billboard.create_or_update_results(rattingbar)

        # Add Corporation Journal to the billboard
        rattingbar_timeline = billboard.create_timeline(self.corporation_journal)
        rattingbar = rattingbar_timeline.annotate_ess_income(is_character=True)
        billboard.create_or_update_results(rattingbar)

        billboard.create_ratting_bar()

        # Aggregate totals
        totals = {
            "total_amount": self.glances.glance_char.aggregate_bounty(self.chars_ids),
            "total_amount_ess": self.glances.glance_corp.aggregate_ess(
                self.chars_ids, is_character=True
            ),
            "total_amount_mining": self.glances.glance_mining.aggregate_mining(
                self.chars_ids
            ),
            "total_amount_costs": self.glances.glance_char.aggregate_costs(
                self.chars_ids
            ),
            "total_amount_others": (
                self.glances.glance_char.aggregate_miscellaneous(self.chars_ids)
                + self.glances.glance_corp.aggregate_daily_goal(
                    self.chars_ids, is_character=True
                )
                + self.glances.glance_char.aggregate_donation(
                    self.chars_ids, exclude=self.alt_ids
                )
            ),
        }
        totals["total_amount_all"] = (
            totals["total_amount"]
            + totals["total_amount_ess"]
            + totals["total_amount_mining"]
            + totals["total_amount_others"]
            + totals["total_amount_costs"]
        )

        output = {
            "ratting": sorted(list(ratting.values()), key=lambda x: x["main_name"]),
            "billboard": billboard.dict,
            "total": totals,
        }

        return output

    def generate_template(self):
        """Generate the template for the character"""
        information_dict = {}

        # Create the Ledger
        ledger_data = InformationData(
            character=self.main,
            date=self.date,
            view=self.view,
            current_date=self.current_date,
        )

        day_glance = AggregateLedger(self.glances.glance_day_char)
        day_glance_corp = AggregateLedger(self.glances.glance_day_corp)
        day_glance_mining = AggregateMining(self.glances.glance_day_mining)

        amounts = defaultdict(lambda: defaultdict(Decimal))

        amounts["bounty_income"] = {
            "total_amount": self.glances.glance_char.aggregate_bounty(self.chars_ids),
            "total_amount_day": day_glance.aggregate_bounty(self.chars_ids),
        }

        amounts["mission_income"] = {
            "total_amount": self.glances.glance_char.aggregate_mission(self.chars_ids),
            "total_amount_day": day_glance.aggregate_mission(self.chars_ids),
        }

        amounts["mining"] = {
            "total_amount": self.glances.glance_mining.aggregate_mining(self.chars_ids),
            "total_amount_day": day_glance_mining.aggregate_mining(self.chars_ids),
        }

        amounts["incursion_income"] = {
            "total_amount": self.glances.glance_char.aggregate_incursion(
                self.chars_ids
            ),
            "total_amount_day": day_glance.aggregate_incursion(self.chars_ids),
        }

        amounts["market_income"] = {
            "total_amount": self.glances.glance_char.aggregate_market(
                second_party=self.chars_ids
            ),
            "total_amount_day": day_glance.aggregate_market(
                second_party=self.chars_ids
            ),
        }

        amounts["contract_income"] = {
            "total_amount": self.glances.glance_char.aggregate_contract(self.chars_ids),
            "total_amount_day": day_glance.aggregate_contract(self.chars_ids),
        }

        amounts["donation_income"] = {
            "total_amount": self.glances.glance_char.aggregate_donation(
                self.chars_ids, exclude=self.alt_ids
            ),
            "total_amount_day": day_glance.aggregate_donation(
                self.chars_ids, exclude=self.alt_ids
            ),
        }

        amounts["insurance_income"] = {
            "total_amount": self.glances.glance_char.aggregate_insurance(
                self.chars_ids
            ),
            "total_amount_day": day_glance.aggregate_insurance(self.chars_ids),
        }

        amounts["milestone_income"] = {
            "total_amount": self.glances.glance_char.aggregate_milestone_reward(
                self.chars_ids
            ),
            "total_amount_day": day_glance.aggregate_milestone_reward(self.chars_ids),
        }

        amounts["market_cost"] = {
            "total_amount": self.glances.glance_char.aggregate_market_cost(
                self.chars_ids
            ),
            "total_amount_day": day_glance.aggregate_market_cost(self.chars_ids),
        }

        amounts["contract_cost"] = {
            "total_amount": self.glances.glance_char.aggregate_contract_cost(
                self.chars_ids
            ),
            "total_amount_day": day_glance.aggregate_contract_cost(self.chars_ids),
        }

        amounts["lp_cost"] = {
            "total_amount": self.glances.glance_char.aggregate_lp_cost(self.chars_ids),
            "total_amount_day": day_glance.aggregate_lp_cost(self.chars_ids),
        }

        amounts["production_cost"] = {
            "total_amount": self.glances.glance_char.aggregate_production_cost(
                self.chars_ids
            ),
            "total_amount_day": day_glance.aggregate_production_cost(self.chars_ids),
        }

        amounts["traveling_cost"] = {
            "total_amount": self.glances.glance_char.aggregate_traveling(
                self.chars_ids
            ),
            "total_amount_day": day_glance.aggregate_traveling(self.chars_ids),
        }

        amounts["asset_cost"] = {
            "total_amount": self.glances.glance_char.aggregate_assets(self.chars_ids),
            "total_amount_day": day_glance.aggregate_assets(self.chars_ids),
        }

        amounts["skill_cost"] = {
            "total_amount": self.glances.glance_char.aggregate_skill(self.chars_ids),
            "total_amount_day": day_glance.aggregate_skill(self.chars_ids),
        }

        amounts["insurance_cost"] = {
            "total_amount": self.glances.glance_char.aggregate_insurance_cost(
                self.chars_ids
            ),
            "total_amount_day": day_glance.aggregate_insurance_cost(self.chars_ids),
        }

        amounts["planetary_cost"] = {
            "total_amount": self.glances.glance_char.aggregate_planetary(
                self.chars_ids
            ),
            "total_amount_day": day_glance.aggregate_planetary(self.chars_ids),
        }

        # Corporation Stuff

        amounts["ess_income"] = {
            "total_amount": self.glances.glance_corp.aggregate_ess(
                self.chars_ids, is_character=True
            ),
            "total_amount_day": day_glance_corp.aggregate_ess(
                self.chars_ids, is_character=True
            ),
        }

        amounts["daily_goal_income"] = {
            "total_amount": self.glances.glance_corp.aggregate_daily_goal(
                self.chars_ids, is_character=True
            ),
            "total_amount_day": day_glance_corp.aggregate_daily_goal(
                self.chars_ids, is_character=True
            ),
        }

        information_dict.update(
            {
                "main_id": ledger_data.id,
                "main_name": ledger_data.name,
                "date": ledger_data.information_date,
            }
        )
        information_dict = ledger_data._generate_amounts_dict(amounts, information_dict)
        return information_dict



import logging
from collections import defaultdict

from django.db import models
from django.db.models import DecimalField, ExpressionWrapper, F, Q, Sum, Value
from django.db.models.functions import Coalesce, Round
from django.utils import timezone

from allianceauth.eveonline.models import EveCharacter

logger = logging.getLogger(__name__)


class AuditCharacterQuerySet(models.QuerySet):
    # pylint: disable=duplicate-code
    def visible_to(self, user):
        # superusers get all visible
        if user.is_superuser:
            logger.debug("Returning all characters for superuser %s.", user)
            return self

        if user.has_perm("ledger.char_audit_admin_manager"):
            logger.debug("Returning all characters for %s.", user)
            return self

        try:
            char = user.profile.main_character
            assert char
            queries = [models.Q(character__character_ownership__user=user)]

            if user.has_perm("ledger.char_audit_manager"):
                queries.append(models.Q(character__corporation_id=char.corporation_id))

            logger.debug(
                "%s queries for user %s visible chracters.", len(queries), user
            )

            query = queries.pop()
            for q in queries:
                query |= q
            return self.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return self.none()


class AuditCharacterManager(models.Manager):
    def get_queryset(self):
        return AuditCharacterQuerySet(self.model, using=self._db)

    @staticmethod
    def visible_eve_characters(user):
        qs = EveCharacter.objects.get_queryset()
        if user.is_superuser:
            logger.debug("Returning all characters for superuser %s.", user)
            return qs.all()

        if user.has_perm("ledger.char_audit_admin_manager"):
            logger.debug("Returning all characters for %s.", user)
            return qs.all()

        try:
            char = user.profile.main_character
            assert char
            queries = [models.Q(character_ownership__user=user)]

            if user.has_perm("ledger.char_audit_manager"):
                queries.append(models.Q(corporation_id=char.corporation_id))

            logger.debug(
                "%s queries for user %s visible chracters.", len(queries), user
            )

            query = queries.pop()
            for q in queries:
                query |= q
            return qs.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return qs.none()

    def visible_to(self, user):
        return self.get_queryset().visible_to(user)


class CharacterMiningLedgerEntryQueryset(models.QuerySet):
    def annotate_pricing(self) -> models.QuerySet:
        """Annotate price and total columns."""
        return self.annotate(price=F("type__market_price__average_price")).annotate(
            total=ExpressionWrapper(
                F("type__market_price__average_price") * F("quantity"),
                output_field=models.DecimalField(),
            ),
        )

    def annotate_mining(self) -> models.QuerySet:
        """Annotate mining columns."""
        return (
            self.annotate_pricing()
            .values(
                "character__character__character_id",
                "character__character__character_name",
            )
            .annotate(
                total_amount=Round(
                    Coalesce(
                        Sum(F("total")),
                        Value(0),
                        output_field=DecimalField(),
                    ),
                    precision=2,
                )
            )
        )

    def aggregate_amounts_information_modal(
        self, amounts: defaultdict, chars_list: list, filter_date: timezone.datetime
    ) -> dict:
        """Generate data template for the ledger character information view."""
        qs = self.filter(Q(character__character__character_id__in=chars_list))
        qs = qs.annotate_pricing()
        qs = qs.aggregate(
            total_amount=Round(
                Coalesce(
                    Sum(F("total"), filter=Q(date__year=filter_date.year)),
                    Value(0),
                    output_field=DecimalField(),
                ),
                precision=2,
            ),
            total_amount_day=Round(
                Coalesce(
                    Sum(F("total"), filter=Q(date__day=filter_date.day)),
                    Value(0),
                    output_field=DecimalField(),
                ),
                precision=2,
            ),
        )

        amounts["mining"]["total_amount"] = qs["total_amount"]
        amounts["mining"]["total_amount_day"] = qs["total_amount_day"]

        return amounts

    def annotate_billboard(self, chars_list: list) -> models.QuerySet:
        """Annotate billboard columns."""
        qs = self.filter(Q(character__character__character_id__in=chars_list))
        return qs.annotate(
            total_amount=Round(
                Coalesce(
                    Sum(F("total")),
                    Value(0),
                    output_field=DecimalField(),
                ),
                precision=2,
            )
        )


class CharacterMiningLedgerEntryManagerBase(models.Manager):
    pass


CharacterMiningLedgerEntryManager = CharacterMiningLedgerEntryManagerBase.from_queryset(
    CharacterMiningLedgerEntryQueryset
)



from unittest.mock import PropertyMock, patch

from django.test import RequestFactory, TestCase
from django.utils import timezone

from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import create_user_from_evecharacter

from ledger.app_settings import LEDGER_CHAR_MAX_INACTIVE_DAYS
from ledger.models.characteraudit import (
    CharacterAudit,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth

MODULE_PATH = "ledger.models.characteraudit"


class TestCharacterWalletJournalModel(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()

        cls.audit = CharacterAudit(
            character=EveCharacter.objects.get(character_id=1001)
        )

    def test_str(self):
        self.assertEqual(str(self.audit), "Gneuten's Character Data")

    def test_is_active_should_true(self):
        self.audit.last_update_wallet = timezone.now()
        self.audit.last_update_mining = timezone.now()
        self.audit.last_update_planetary = timezone.now()
        self.assertTrue(self.audit.is_active())

    def test_is_active_should_false(self):
        self.audit.last_update_wallet = timezone.now() - timezone.timedelta(days=4)
        self.audit.last_update_mining = timezone.now() - timezone.timedelta(days=4)
        self.audit.last_update_planetary = timezone.now() - timezone.timedelta(days=4)
        self.assertFalse(self.audit.is_active())

    def test_is_active_exception(self):
        self.audit.last_update_wallet = None
        self.audit.last_update_mining = None
        self.audit.last_update_planetary = None
        self.assertFalse(self.audit.is_active())

    @patch(MODULE_PATH + ".logger")
    def test_is_active_should_deactive_character(self, mock_logger):
        self.audit.active = True
        self.audit.last_update_wallet = timezone.now()
        self.audit.last_update_mining = timezone.now() - timezone.timedelta(days=4)
        self.audit.last_update_planetary = timezone.now() - timezone.timedelta(days=4)
        self.assertFalse(self.audit.is_active())
        mock_logger.info.assert_called_once_with(
            "Deactivating Character: %s", self.audit.character.character_name
        )

    def test_get_esi_scopes(self):
        self.assertEqual(
            self.audit.get_esi_scopes(),
            [
                # Mining Ledger
                "esi-industry.read_character_mining.v1",
                # Wallet / Market /  Contracts
                "esi-wallet.read_character_wallet.v1",
                "esi-contracts.read_character_contracts.v1",
                # Planetary Interaction
                "esi-planets.manage_planets.v1",
            ],
        )

    def test_get_status_opacity_should_return_100(self):
        self.audit.active = True
        self.assertEqual(self.audit.get_status_opacity, "opacity-100")

    def test_get_status_opacity_should_return_25(self):
        self.audit.active = False
        self.assertEqual(self.audit.get_status_opacity, "opacity-25")

    def test_get_status_icon_should_return_ok(self):
        self.audit.last_update_mining = timezone.now()
        self.audit.last_update_wallet = timezone.now()
        self.audit.last_update_planetary = timezone.now()
        self.audit.active = True

        self.assertEqual(self.audit.get_status, self.audit.UpdateStatus("ok"))
        self.assertEqual(
            self.audit.get_status.bootstrap_icon(),
            self.audit.UpdateStatus("ok").bootstrap_icon(),
        )

    def test_get_status_icon_should_return_disabled(self):
        self.audit.active = False
        self.assertEqual(
            self.audit.get_status,
            self.audit.UpdateStatus("disabled"),
        )

    def test_get_status_icon_should_return_not_up_to_date(self):
        self.audit.active = True
        self.audit.last_update_mining = timezone.now() - timezone.timedelta(days=4)
        self.audit.last_update_wallet = timezone.now()
        self.audit.last_update_planetary = timezone.now()
        self.assertEqual(
            self.audit.get_status,
            self.audit.UpdateStatus("not_up_to_date"),
        )



"""
Character Audit Model
"""

import datetime
import logging

from django.db import models
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from eveuniverse.models import EveSolarSystem, EveType

from allianceauth.eveonline.models import EveCharacter

from ledger import app_settings
from ledger.managers.characterjournal_manager import CharWalletManager
from ledger.managers.charaudit_manager import (
    AuditCharacterManager,
    CharacterMiningLedgerEntryManager,
)
from ledger.models.general import EveEntity

logger = logging.getLogger(__name__)


class CharacterAudit(models.Model):
    class UpdateStatus(models.TextChoices):
        DISABLED = "disabled", _("disabled")
        NOT_UP_TO_DATE = "not_up_to_date", _("not up to date")
        ERROR = "error", _("error")
        OK = "ok", _("ok")

        def bootstrap_icon(self) -> str:
            """Return bootstrap corresponding icon class."""
            update_map = {
                self.DISABLED: mark_safe(
                    f"<span class='{self.bootstrap_text_style_class()}' data-tooltip-toggle='ledger-tooltip' title='{self.description()}'>⬤</span>"
                ),
                self.NOT_UP_TO_DATE: mark_safe(
                    f"<span class='{self.bootstrap_text_style_class()}' data-tooltip-toggle='ledger-tooltip' title='{self.description()}'>⬤</span>"
                ),
                self.ERROR: mark_safe(
                    f"<span class='{self.bootstrap_text_style_class()}' data-tooltip-toggle='ledger-tooltip' title='{self.description()}'>⬤</span>"
                ),
                self.OK: mark_safe(
                    f"<span class='{self.bootstrap_text_style_class()}' data-tooltip-toggle='ledger-tooltip' title='{self.description()}'>⬤</span>"
                ),
            }
            return update_map.get(self, "")

        def bootstrap_text_style_class(self) -> str:
            """Return bootstrap corresponding bootstrap text style class."""
            update_map = {
                self.DISABLED: "text-danger",
                self.NOT_UP_TO_DATE: "text-warning",
                self.ERROR: "text-danger",
                self.OK: "text-success",
            }
            return update_map.get(self, "")

        def description(self) -> str:
            """Return description for an enum object."""
            update_map = {
                self.DISABLED: _("Update is disabled"),
                self.NOT_UP_TO_DATE: _(
                    "One of the updates is older than {} day"
                ).format(app_settings.LEDGER_CHAR_MAX_INACTIVE_DAYS),
                self.ERROR: _("An error occurred during update"),
                self.OK: _("Updates completed successfully"),
            }
            return update_map.get(self, "")

    id = models.AutoField(primary_key=True)

    character_name = models.CharField(max_length=100, null=True, default=None)

    character = models.OneToOneField(
        EveCharacter, on_delete=models.CASCADE, related_name="ledger_character"
    )

    active = models.BooleanField(default=False)

    last_update_wallet = models.DateTimeField(null=True, default=None, blank=True)

    last_update_mining = models.DateTimeField(null=True, default=None, blank=True)

    last_update_planetary = models.DateTimeField(null=True, default=None, blank=True)

    balance = models.DecimalField(
        max_digits=20, decimal_places=2, null=True, default=None
    )

    objects = AuditCharacterManager()

    def __str__(self):
        return f"{self.character.character_name}'s Character Data"

    class Meta:
        default_permissions = ()
        permissions = (
            ("char_audit_manager", "Has access to all characters for own Corp"),
            ("char_audit_admin_manager", "Has access to all characters"),
        )

    @classmethod
    def get_esi_scopes(cls) -> list[str]:
        """Return list of required ESI scopes to fetch."""
        return [
            # Mining Ledger
            "esi-industry.read_character_mining.v1",
            # Wallet / Market /  Contracts
            "esi-wallet.read_character_wallet.v1",
            "esi-contracts.read_character_contracts.v1",
            # Planetary Interaction
            "esi-planets.manage_planets.v1",
        ]

    def _get_is_active(self):
        time_ref = timezone.now() - datetime.timedelta(
            days=app_settings.LEDGER_CHAR_MAX_INACTIVE_DAYS
        )
        try:
            is_active = True

            is_active = is_active and self.last_update_wallet > time_ref
            is_active = is_active and self.last_update_mining > time_ref
            is_active = is_active and self.last_update_planetary > time_ref
            return is_active
        except Exception:  # pylint: disable=broad-exception-caught
            return False

    def is_active(self):
        is_active = self._get_is_active()
        if self.active != is_active:
            self.active = is_active
            if is_active is False:
                logger.info("Deactivating Character: %s", self.character.character_name)
            self.save()
        return is_active

    # TODO Create a Update Status Model to handle all update section separately
    @property
    def get_status(self):
        if self.active is False:
            return self.UpdateStatus("disabled")

        is_active = self._get_is_active()

        # Check if one of the updates are not up to date
        if is_active is False:
            return self.UpdateStatus("not_up_to_date")
        return self.UpdateStatus("ok")

    @property
    def get_status_opacity(self):
        if self.active:
            return "opacity-100"
        return "opacity-25"


class WalletJournalEntry(models.Model):
    amount = models.DecimalField(
        max_digits=20, decimal_places=2, null=True, default=None
    )
    balance = models.DecimalField(
        max_digits=20, decimal_places=2, null=True, default=None
    )
    context_id = models.BigIntegerField(null=True, default=None)

    class ContextType(models.TextChoices):
        STRUCTURE_ID = "structure_id"
        STATION_ID = "station_id"
        MARKET_TRANSACTION_ID = "market_transaction_id"
        CHARACTER_ID = "character_id"
        CORPORATION_ID = "corporation_id"
        ALLIANCE_ID = "alliance_id"
        EVE_SYSTEM = "eve_system"
        INDUSTRY_JOB_ID = "industry_job_id"
        CONTRACT_ID = "contract_id"
        PLANET_ID = "planet_id"
        SYSTEM_ID = "system_id"
        TYPE_ID = "type_id"

    context_id_type = models.CharField(
        max_length=30, choices=ContextType.choices, null=True, default=None
    )
    date = models.DateTimeField()
    description = models.CharField(max_length=500)
    first_party = models.ForeignKey(
        EveEntity,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        related_name="+",
    )
    entry_id = models.BigIntegerField()
    reason = models.CharField(max_length=500, null=True, default=None)
    ref_type = models.CharField(max_length=72)
    second_party = models.ForeignKey(
        EveEntity,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        related_name="+",
    )
    tax = models.DecimalField(max_digits=20, decimal_places=2, null=True, default=None)
    tax_receiver_id = models.IntegerField(null=True, default=None)

    class Meta:
        abstract = True
        indexes = (
            models.Index(fields=["date"]),
            models.Index(fields=["amount"]),
            models.Index(fields=["entry_id"]),
            models.Index(fields=["ref_type"]),
            models.Index(fields=["first_party"]),
            models.Index(fields=["second_party"]),
        )
        default_permissions = ()


class CharacterWalletJournalEntry(WalletJournalEntry):
    character = models.ForeignKey(CharacterAudit, on_delete=models.CASCADE)

    objects = CharWalletManager()

    def __str__(self):
        return f"Character Wallet Journal: RefType: {self.ref_type} - {self.first_party.name} -> {self.second_party.name}: {self.amount} ISK"

    @classmethod
    def get_visible(cls, user):
        chars_vis = CharacterAudit.objects.visible_to(user)
        return cls.objects.filter(character__in=chars_vis)


# Mining Models


class CharacterMiningLedger(models.Model):
    id = models.CharField(max_length=50, primary_key=True)
    character = models.ForeignKey(
        CharacterAudit, on_delete=models.CASCADE, related_name="ledger_character"
    )
    date = models.DateTimeField()
    type = models.ForeignKey(
        EveType, on_delete=models.CASCADE, related_name="ledger_evetype"
    )
    system = models.ForeignKey(
        EveSolarSystem, on_delete=models.CASCADE, related_name="ledger_evesolarsystem"
    )
    quantity = models.IntegerField()

    @staticmethod
    def create_primary_key(character_id, mining_record):
        return f"{mining_record['date'].strftime('%Y%m%d')}-{mining_record['type_id']}-{character_id}-{mining_record['solar_system_id']}"

    objects = CharacterMiningLedgerEntryManager()

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        return f"{self.character} {self.id}"
